/*********************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 * 
 * --SA4.1_PID--
 * Copyright IBM Corp. 2005, 2006, 2007, 2008
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what
 * been deposited with the U.S. Copyright Office. 
 *-----------------------------------------------				
 * Author: Surren Partabh
 * Created: 06/01/2005
 * Description:
 *   Script to create the administration package bodies
 *-----------------------------------------------				
 * MODIFICATION HISTORY
 ********************************************************************/

CREATE OR REPLACE PACKAGE BODY lock_util
--
-- @(#) lock_util.sql 1.1 0 07/04/05 13:51:15
-- @(#)
IS
  PROCEDURE check_result(lockName IN VARCHAR2, result IN INTEGER);
  PROCEDURE get_lock(lockName IN VARCHAR2, lockType IN INTEGER);


  PROCEDURE get_x(lockName IN VARCHAR2)
  IS
  BEGIN
    get_lock(lockName, dbms_lock.x_mode);
  END get_x;

  PROCEDURE release(lockName IN VARCHAR2)
  IS
    lockHandle  VARCHAR2(200);
    result      PLS_INTEGER;
  BEGIN
    dbms_lock.allocate_unique(lockName, lockHandle);

    result := dbms_lock.release(lockHandle);
    check_result(lockName, result);
  END release;

  PROCEDURE get_r(lockName IN VARCHAR2)
  IS
  BEGIN
    get_lock(lockName, dbms_lock.ss_mode);
  END get_r;


  PROCEDURE get_lock(lockName IN VARCHAR2, lockType IN INTEGER)
  IS
    lockHandle  VARCHAR2(200);
    result      PLS_INTEGER;
  BEGIN
    dbms_lock.allocate_unique(lockName, lockHandle);
    result := dbms_lock.request(lockHandle, lockType);
    check_result(lockName, result);
  END get_lock;


  PROCEDURE check_result(lockName IN VARCHAR2, result IN INTEGER)
  IS
  BEGIN
    IF (result <> 0)
    THEN
      dbms_output.put_line('Error acquiring lock: ' || lockName || ': ' ||
           case
              when result=1 then 'Timeout'
              when result=2 then 'Deadlock'
              when result=3 then 'Parameter Error'
              when result=4 then 'Already owned'
              when result=5 then 'Illegal Lock Handle'
            END);
    END IF;
  END check_result;


END lock_util;
/
SHOW ERRORS PACKAGE BODY lock_util;

