/*********************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 * 
 * --SA4.1_PID--
 * Copyright IBM Corp. 2005, 2006, 2007, 2008
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what
 * been deposited with the U.S. Copyright Office. 
 *-----------------------------------------------				
 * Author: Surren Partabh
 * Created: 06/01/2005
 * Description:
 *   Script to create the administration package bodies
 *-----------------------------------------------				
 * MODIFICATION HISTORY
 ********************************************************************/

PROMPT Creating Package Body 'LOG_PKG'

CREATE OR REPLACE PACKAGE BODY log_pkg
IS

    /* 
     * initialised to count(*) of the configuration
     * when 0 the java log4j code is not invoked
     */
    java_conf NUMBER(6):=0;
    java_debug_conf NUMBER(6):=0;
    trace_directory VARCHAR(2000);

    /*
     * LOG4J Wrappers
     */
    PROCEDURE log4j_warn(category IN VARCHAR2, code IN VARCHAR2,
        message IN VARCHAR2, event IN VARCHAR2) IS
        LANGUAGE JAVA NAME 'com.comnitel.oracle.log.embedded.OracleLogger.warn(
            java.lang.String,java.lang.String,java.lang.String,java.lang.String)';
    PROCEDURE log4j_error(category IN VARCHAR2, code IN VARCHAR2,
        message IN VARCHAR2, event IN VARCHAR2) IS
        LANGUAGE JAVA NAME 'com.comnitel.oracle.log.embedded.OracleLogger.error(
            java.lang.String,java.lang.String,java.lang.String,java.lang.String)';
    PROCEDURE log4j_debug(category IN VARCHAR2, message IN VARCHAR2) IS
        LANGUAGE JAVA NAME
        'com.comnitel.oracle.log.embedded.OracleLogger.debug(
            java.lang.String,java.lang.String)';
    PROCEDURE log4j_info(category IN VARCHAR2, message IN VARCHAR2) IS
        LANGUAGE JAVA NAME
        'com.comnitel.oracle.log.embedded.OracleLogger.info(
            java.lang.String,java.lang.String)';
    FUNCTION log4j_is_debug_on(category IN VARCHAR2) RETURN NUMBER IS
        LANGUAGE JAVA NAME
        'com.comnitel.oracle.log.embedded.OracleLogger.isDebugEnabled(
            java.lang.String) return int';
    PROCEDURE log4j_reconfigure IS
      LANGUAGE JAVA NAME
        'com.comnitel.oracle.log.embedded.OracleLogger.reconfigure()';

    -- Log to file
    PROCEDURE log_to_file(message IN VARCHAR2) IS
        l_date DATE;
        file utl_file.file_type;
    BEGIN
        IF trace_directory IS NULL THEN
            RETURN;
        END IF;
        l_date:=sysdate;
        file:=utl_file.fopen(trace_directory,
            'wmc-'||user||'-'||to_char(l_date,'yyyymmdd')||'.log','a',6000);
        utl_file.put_line(file,to_char(l_date,'yyyy/mm/dd-hh')||':'||message,FALSE);
        utl_file.fclose(file);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Log to file
    PROCEDURE log_to_file(a IN VARCHAR2, b IN VARCHAR2) IS
        l_date DATE;
        file utl_file.file_type;
    BEGIN
        IF trace_directory IS NULL THEN
            RETURN;
        END IF;
        l_date:=sysdate;
        file:=utl_file.fopen(trace_directory,
            'wmc-'||user||'-'||to_char(l_date,'yyyymmdd')||'.log','a',6000);
        utl_file.put_line(file,to_char(l_date,'Dy DD-Mon-YYYY HH24:MI:SS')||':'||a||':'||b,FALSE);
        utl_file.fclose(file);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Log to file
    PROCEDURE log_to_file(a IN VARCHAR2, b IN NUMBER, c IN VARCHAR2, d IN VARCHAR2) IS
        l_date DATE;
        file utl_file.file_type;
    BEGIN
        IF trace_directory IS NULL THEN
            RETURN;
        END IF;
        l_date:=sysdate;
        file:=utl_file.fopen(trace_directory,
            'wmc-'||user||'-'||to_char(l_date,'yyyymmdd')||'.log','a',6000);
        utl_file.put_line(file,to_char(l_date,'Dy DD-Mon-YYYY HH24:MI:SS')||':'||a||
            ':'||b||':'||c||':'||d,FALSE);
        utl_file.fclose(file);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a debug message
    PROCEDURE debug(category IN VARCHAR2, message IN VARCHAR2)
    IS
    BEGIN
        IF java_conf = 0 THEN
            RETURN;
        END IF;
        log4j_debug(category,message);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log an info message
    PROCEDURE info(category IN VARCHAR2, message IN VARCHAR2)
    IS
    BEGIN
        log_to_file(category,message);
        IF java_conf = 0 THEN
            RETURN;
        END IF;
        log4j_info(category,message);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a warning message
    PROCEDURE warn(category IN VARCHAR2, code IN NUMBER DEFAULT SQLCODE,
      message IN VARCHAR2 DEFAULT SQLERRM, event IN VARCHAR2 DEFAULT 'GEN')
    IS
    BEGIN
        log_to_file(category,code,message,event);
        IF java_conf = 0 THEN
            RETURN;
        END IF;
        log4j_warn(category,code,message,event);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a error message, also log into ADM_ERROR_LOG
    PROCEDURE error(category IN VARCHAR2, code IN NUMBER DEFAULT SQLCODE,
      message IN VARCHAR2 DEFAULT SQLERRM, event IN VARCHAR2 DEFAULT 'GEN')
    IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        log_to_file(category,code,message,event);
        INSERT INTO adm_error_log VALUES (event, trim(category),
            sysdate, code, trim(message));
        COMMIT;
        IF java_conf = 0 THEN
            RETURN;
        END IF;
        log4j_error(category,code,message,event);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a message into ADM_LOG
    PROCEDURE log_info_table(l_event_name IN VARCHAR2, l_object_name IN VARCHAR2, l_job_id IN NUMBER,
                             l_task_id IN NUMBER, l_log_time IN DATE, l_message IN VARCHAR2)
    IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        info(l_event_name,l_object_name||' : '||l_message);
        INSERT INTO adm_log (event_name, object_name, job_id, task_id,
                log_time, sql_code, message)
        VALUES (l_event_name, l_object_name, l_job_id, l_task_id,
                l_log_time, NULL, l_message);
        COMMIT;
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a warning message into ADM_LOG and ADM_ERROR_LOG
    PROCEDURE log_warn_table(l_event_name IN VARCHAR2, l_object_name IN VARCHAR2, l_job_id IN NUMBER,
                             l_task_id IN NUMBER, l_log_time IN DATE, l_message IN VARCHAR2,
                             l_sql_code IN NUMBER DEFAULT NULL)
    IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        warn(l_event_name, l_sql_code, l_message, l_object_name);       
        INSERT INTO adm_error_log VALUES (l_event_name, l_object_name,
            sysdate, l_sql_code, trim(l_message));        
        INSERT INTO adm_log (event_name, object_name, job_id, task_id,
                log_time, sql_code, message)
        VALUES (l_event_name, l_object_name, l_job_id, l_task_id,
                l_log_time, l_sql_code, l_message);
        COMMIT;
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- Procedure to log a error message into ADM_ERROR_LOG
    PROCEDURE log_error_table(code IN NUMBER, msg IN VARCHAR2, event IN VARCHAR2,
      evsrc IN VARCHAR2)
    IS
    BEGIN
        error(evsrc,code,msg,event);
    END;

    -- Tracing level check
    FUNCTION is_debug_on(category IN VARCHAR2) RETURN BOOLEAN
    IS
    BEGIN
        IF java_debug_conf = 0 THEN
            RETURN FALSE;
        END IF;
        IF java_conf = 0 THEN
            RETURN FALSE;
        END IF;
        RETURN (log4j_is_debug_on(category)=1);
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            RETURN FALSE;
        END;
    END;

    -- Tracing reconfiguration
    PROCEDURE reconfigure IS
        new_java_conf java_conf%TYPE;
    BEGIN
        SELECT COUNT(*) INTO new_java_conf FROM adm_log_conf
            WHERE name LIKE 'log4j%';
        SELECT COUNT(*) INTO java_debug_conf FROM adm_log_conf
            WHERE name LIKE 'log4j%' AND value LIKE '%DEBUG%';
        SELECT value INTO trace_directory FROM adm_log_conf
            WHERE name='trace.directory';
        IF new_java_conf <> 0 THEN
            java_conf:=new_java_conf;
            log4j_reconfigure;
            RETURN;
        END IF;
        IF java_conf <> 0 THEN
            log4j_reconfigure;
        END IF;
        java_conf:=new_java_conf;
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            NULL;
        END;
    END;

    -- access to trace directory
    FUNCTION get_trace_directory RETURN VARCHAR2 IS
    BEGIN
        return trace_directory;
    END;

    -- Configuration table manipulation
    PROCEDURE set_conf_property(cname IN VARCHAR2, value IN VARCHAR2) IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        DELETE FROM adm_log_conf c WHERE c.name=cname;
        INSERT INTO adm_log_conf VALUES(cname,value);
        COMMIT;
    END;

    -- Configuration table manipulation
    PROCEDURE clear_conf_property(cname IN VARCHAR2) IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        DELETE FROM adm_log_conf c WHERE c.name=cname;
        COMMIT;
    END;

    -- Configuration table manipulation
    PROCEDURE clear_conf_properties IS
        PRAGMA AUTONOMOUS_TRANSACTION;
    BEGIN
        DELETE FROM adm_log_conf;
        COMMIT;
    END;


BEGIN
    SELECT COUNT(*) INTO java_conf FROM adm_log_conf
        WHERE name LIKE 'log4j%';
    SELECT COUNT(*) INTO java_debug_conf FROM adm_log_conf
        WHERE name LIKE 'log4j%' AND value LIKE '%DEBUG%';
    BEGIN
        SELECT value INTO trace_directory FROM adm_log_conf
            WHERE name='trace.directory';
    EXCEPTION
        WHEN OTHERS THEN
        BEGIN
            trace_directory:=null;
        END;
    END;
END log_pkg;
/
SHOW ERROR

-- Set up the trace directory for the log package
EXEC log_pkg.set_conf_property('trace.directory','TRACE_FILE');

