/*********************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 * 
 * --SA4.1_PID--
 * Copyright IBM Corp. 2005, 2006, 2007, 2008
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what
 * been deposited with the U.S. Copyright Office. 
 *-----------------------------------------------                
 * Author: Thierry Supplisson
 * Created: 06/01/2005
 * Description:
 *   Script to create the admin packages
 *-----------------------------------------------                
 * MODIFICATION HISTORY
 * DATE        MODIFIED BY    MODIFICATION
 ********************************************************************/

/* This package allows the logging of errors from 
other procedures/packages/triggers */

/*

  Usage of the log server
  1. Start the log server on a serviceassure host <HOST>:
    $WMCROOT/bin/oracle_logserver <PORT>
  2. Configure the system to communicate with this log server
     (i.e. from sqlplus):
    exec log_pkg.clear_conf_properties;
    exec log_pkg.set_conf_property('log4j.appender.DEF',
        'com.comnitel.oracle.log.embedded.SocketAppender');
    exec log_pkg.set_conf_property('log4j.appender.DEF.Port','<PORT>');
    exec log_pkg.set_conf_property('log4j.appender.DEF.RemoteHost','<HOST>');
  3. For each category CAT you plan to use
     (i.e. from sqlplus):
    exec log_pkg.set_conf_property('log4j.logger.WMC.CAT','DEBUG');
  4. Make sure that your Oracle user is granted the JAVASYSPRIV (this
     should be done BEFORE loading the Java logging classes).
  5. reconfigure the logging:
    exec log_pkg.reconfigure;

  A. if you want to go back to default configuration
    exec log_pkg.clear_conf_properties;
    exec log_pkg.reconfigure;

  B. It is recommended to use hierachical categories, i.e. WMC.BLAH.
     doing so you can set global tracing levels
     exec log_pkg.set_conf_property('log4j.logger.XXX','DEBUG'); or
     exec log_pkg.set_conf_property('log4j.logger.XXX.BLAH','DEBUG');
  C. The top category is WMC prepended to any logging call. i.e.
        log_pkg.error('TOTO') logs as 'WMC.TOTO';
  D. The configuration is stored in the table adm_log_conf

  E. If the logserver is stopped and restarted, it is necessary to
     run reconfigure in the database.

  D. trace.directory in adm_log_conf contain the file trace directory object:
     create or replace directory TRACE_FILE as 'c:\temp'
     exec log_pkg.set_conf_property('trace.directory','TRACE_FILE');

 */
PROMPT Creating Package 'LOG_PKG'
CREATE OR REPLACE PACKAGE log_pkg
IS
    /*
     * Procedure to reconfigure the logging API
     */
    PROCEDURE reconfigure;

    /*
     * Procedure to send a message to the log server
     */
    PROCEDURE debug(category IN VARCHAR2, message IN VARCHAR2);
    PROCEDURE info(category IN VARCHAR2, message IN VARCHAR2);
    PROCEDURE warn(category IN VARCHAR2, code IN NUMBER DEFAULT SQLCODE,
        message IN VARCHAR2 DEFAULT SQLERRM, event IN VARCHAR2 DEFAULT 'GEN');
    PROCEDURE error(category IN VARCHAR2, code IN NUMBER DEFAULT SQLCODE,
        message IN VARCHAR2 DEFAULT SQLERRM, event IN VARCHAR2 DEFAULT 'GEN');
        
    PROCEDURE log_info_table(l_event_name IN VARCHAR2, l_object_name IN VARCHAR2, l_job_id IN NUMBER, 
                             l_task_id IN NUMBER, l_log_time IN DATE, l_message IN VARCHAR2);

    PROCEDURE log_warn_table(l_event_name IN VARCHAR2, l_object_name IN VARCHAR2, l_job_id IN NUMBER, 
                             l_task_id IN NUMBER, l_log_time IN DATE, l_message IN VARCHAR2, 
                             l_sql_code IN NUMBER DEFAULT NULL);

    PROCEDURE log_error_table(code IN NUMBER, msg IN VARCHAR2,
        event IN VARCHAR2, evsrc IN VARCHAR2);

    /*
     * Test configuration
     */
    FUNCTION is_debug_on(category IN VARCHAR2) RETURN BOOLEAN;

    /*
     * Interact with the configuration
     */
    PROCEDURE set_conf_property(cname IN VARCHAR2, value IN VARCHAR2);
    PROCEDURE clear_conf_property(cname IN VARCHAR2);
    PROCEDURE clear_conf_properties;

    FUNCTION get_trace_directory RETURN VARCHAR2;
END;
/
SHOW ERROR
