/*********************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 * 
 * --SA4.1_PID--
 * Copyright IBM Corp. 2005, 2006, 2007, 2008
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what
 * been deposited with the U.S. Copyright Office. 
 *-----------------------------------------------				
 * Author: Surren Partabh
 * Created: 06/01/2005
 * Description:
 *   Script to create the administration package bodies
 *-----------------------------------------------				
 * MODIFICATION HISTORY
 ********************************************************************/

CREATE OR REPLACE PACKAGE BODY
  -- -----------------------------------------------------------------------
  --
  -- Copyright IBM Corp. 2007
  --
  -- @(#) part_maint.sql 1.54.0 07/04/12 15:44:42
  -- @(#)
  -- -----------------------------------------------------------------------
 Part_Maint IS

   g_tablespace_name    VARCHAR2(30)    := NULL;
   g_extent_size        INTEGER         := NULL;

   g_task_lock_name     VARCHAR2(30)    := 'part-maint-get-next-task-lock';

   g_max_run_time       INTEGER := 60*60*6;      -- Total max time to do p_m stuff
   g_analyze_pct        INTEGER := 10;           -- Estimate pctage for DBMS_STATS
   g_update_move_pct    INTEGER := 20;           -- % rows updated before move/analyze
   g_stats_parallel     INTEGER := 2;            -- parallel statistics

   g_poll_interval      INTEGER := 10;           -- poll interval in seconds
   g_retry_interval     INTEGER := 5;            -- poll interval in seconds

   g_resource_attempts  INTEGER := 10;           -- resource_busy attempts
   g_resource_interval  INTEGER := 5;            -- resource_busy interval

   g_min_part_sessions    INTEGER := 1;         -- This is the MIN number of sessions
   g_part_sessions        INTEGER := 1;         -- This is the config number of sessions

   -- OP values and default weights
   g_Drop_Op            INTEGER := 1;
   g_Add_Op             INTEGER := 2;
   g_Split_Op           INTEGER := 3;
   g_Move_Op            INTEGER := 4;           -- Not used due to Oracle issues
   g_Analyze_Op         INTEGER := 5;           -- Not used due to Oracle issues
   g_MoveAnalyze_Op     INTEGER := 6;           -- Combined to get around Oracle issues on Move and Analyze

   -- OP status values.
   g_OPS_ABORT     CONSTANT INTEGER := 1;         -- We have been interrupted, ABORT
   g_OPS_READY     CONSTANT INTEGER := 2;         -- Waiting for a part_task
   g_OPS_ACTIVE    CONSTANT INTEGER := 3;         -- part_task is processing
   g_OPS_SUCCESS   CONSTANT INTEGER := 4;         -- Success
   g_OPS_FAILED    CONSTANT INTEGER := 5;         -- Failed
   g_OPS_INTERRUPT CONSTANT INTEGER := 8;         -- Interrupted

   pm_failure_c  CONSTANT INTEGER := 0;         -- FALSE
   pm_success_c  CONSTANT INTEGER := 1;         -- TRUE

   g_module_name CONSTANT VARCHAR2(30) := 'PartMaint'; -- DBMS_Application_Info string

   -- Global to hold the session ids
   session_id_tab session_id_tabletype := session_id_tabletype();

   -- -----------------------------------------------------------------------
   --  Private functions/procedures
   -- -----------------------------------------------------------------------
   FUNCTION maintainall(period_in CHAR)  RETURN INTEGER;

   FUNCTION cleanbehind( table_desc taskRec ) RETURN BOOLEAN;

   FUNCTION pushahead( table_desc taskRec ) RETURN BOOLEAN;

   FUNCTION pushbehind( table_desc taskRec ) RETURN BOOLEAN;

   FUNCTION movepartitions( table_desc taskRec ) RETURN BOOLEAN;

   FUNCTION analyzepartitions( table_desc taskRec ) RETURN BOOLEAN;

   PROCEDURE check_unmanaged_part_tables(jobidparam IN INTEGER) ;

   FUNCTION getextentsize( tablespace_name_in VARCHAR2 ) RETURN INTEGER;

   PROCEDURE init_tunable_parameters;

   FUNCTION job_interrupted(job_id_in INTEGER) RETURN BOOLEAN;

   PROCEDURE update_part_task( p_jobid IN INTEGER, p_taskid IN INTEGER, p_code IN INTEGER);

   PROCEDURE update_ready_count( p_jobid IN INTEGER );

   PROCEDURE create_ad_ops( p_period IN CHAR, p_eff_date IN DATE, p_jobid IN INTEGER );

   PROCEDURE create_split_ops( p_period IN CHAR, p_eff_date IN DATE, p_jobid IN INTEGER );

   PROCEDURE create_ma_ops(startdate IN DATE, enddate IN DATE, period_in IN CHAR,
                          usetabmodifications IN BOOLEAN DEFAULT FALSE,
                          jobidparam IN INTEGER DEFAULT NULL);

   PROCEDURE create_add_task( p_jobid IN INTEGER, p_eff_date IN DATE, p_task IN taskRec );

   PROCEDURE create_drop_task( p_jobid IN INTEGER, p_eff_date IN DATE, p_task IN taskRec );

   PROCEDURE create_split_task( p_jobid IN INTEGER, p_eff_date IN DATE, p_task IN taskRec );

   PROCEDURE create_ma_task( p_jobid IN INTEGER, p_eff_date IN DATE, p_task IN taskRec );

   PROCEDURE create_part_sessions( p_jobid IN INTEGER, p_part_sessions IN INTEGER );

   PROCEDURE abort_tasks( p_jobid IN INTEGER );

   FUNCTION all_sessions_complete RETURN BOOLEAN;

   FUNCTION all_tasks_success( p_jobid IN INTEGER ) RETURN BOOLEAN;

   FUNCTION waiting_tasks( p_jobid IN INTEGER ) RETURN BOOLEAN;

   PROCEDURE get_next_task( p_jobid IN INTEGER, p_task IN OUT NOCOPY taskRec );

   FUNCTION task_poll( jobidparam IN INTEGER) RETURN INTEGER;

   FUNCTION generic_start_task(job_id INTEGER, msg VARCHAR2) RETURN INTEGER;

   PROCEDURE generic_end_task(job_id INTEGER, task_id INTEGER,
			      msg VARCHAR2, success BOOLEAN DEFAULT TRUE);

   -- -----------------------------------------------------------------------
   -- Maintain* - Do all part maint operations (clean behind, push ahead,
   --           push behind, move, analyze) for the effective date of the job.
   -- -----------------------------------------------------------------------

   PROCEDURE maintainminly IS
      l_return INTEGER;
   BEGIN
       l_return := maintainall('Q');
       RETURN;
   END maintainminly;

   -- -----------------------------------------------------------------------
   PROCEDURE maintainhourly IS
      l_return INTEGER;
   BEGIN
       l_return := maintainall('H');
       RETURN;
   END maintainhourly;

   -- -----------------------------------------------------------------------
   PROCEDURE maintaindaily IS
      l_return INTEGER;
   BEGIN
      l_return := maintainall('D');
      RETURN;
   END maintaindaily;

   -- -----------------------------------------------------------------------
   PROCEDURE maintainweekly IS
      l_return INTEGER;
   BEGIN
      l_return := maintainall('W');
      RETURN;
   END maintainweekly;

   -- -----------------------------------------------------------------------
   PROCEDURE maintainmonthly IS
      l_return INTEGER;
   BEGIN
      l_return := maintainall('M');
      RETURN;
   END maintainmonthly;

   -- -----------------------------------------------------------------------
   -- MaintainAll - The main interface to partition maintenance.
   -- Loop through all the tables in part_tables and perform all of the
   -- partition maintenace on one table.  We do all of the ops on one table
   -- at a time to minimize the amount of free space we need and hopefully
   -- take advantage of some caching.
   -- -----------------------------------------------------------------------
   FUNCTION maintainall(period_in CHAR) RETURN INTEGER IS
      jobidparam    INTEGER;
      success       BOOLEAN      := TRUE;
      eff_date      DATE;
      vprc          VARCHAR2(30) := 'MaintainAll';
      vreturn       INTEGER      := pm_success_c;

   BEGIN

      dbms_application_info.set_module( g_module_name, NULL );
      dbms_application_info.set_client_info( 'Started' );

            -- setup a unique job id for the partition maintenance job
      SELECT seq_pm_job.NEXTVAL INTO jobidparam FROM dual;

      Log_Pkg.log_info_table(g_module_name, vprc, jobidparam, NULL,
                            SYSDATE, 'Partition MaintainAll Started for period '||period_in);

      init_tunable_parameters();

      create_part_sessions( jobidparam, g_part_sessions );
      create_ad_ops( period_in, SYSDATE, jobidparam );
      create_ma_ops( NULL, NULL, period_in, TRUE, jobidparam );

      -- Commit to persist the operations and start the jobs
      COMMIT;

      update_ready_count( jobidparam );

      -- Poll while the sessions are running
      vreturn := task_poll( jobidparam );

      Log_Pkg.log_info_table(g_module_name, vprc, jobidparam, NULL,
                            SYSDATE, 'Partition MaintainAll Completed for period '||period_in);

      RETURN vreturn;

   EXCEPTION
      WHEN OTHERS THEN
         Log_Pkg.log_warn_table(g_module_name, vprc, jobidparam, NULL, SYSDATE,
                            'Exception Partition Maintanance failed: '||SQLERRM, SQLCODE);
         RETURN pm_failure_c;

   END maintainall;

   -- -----------------------------------------------------------------------
   -- CleanBehind - Drop partitions that are too old and not in the
   -- PART_PINNED table.
   -- -----------------------------------------------------------------------
   FUNCTION cleanbehind( table_desc taskRec ) RETURN BOOLEAN IS
      retval      BOOLEAN := TRUE;
      retmsg      VARCHAR2(9000) := NULL;
      sqlstr      VARCHAR2(4000);
      delete_date DATE;

      CURSOR droppartition(table_name_in VARCHAR2,
             table_id_in INTEGER, delete_date DATE) IS
            SELECT *
            FROM user_objects t, PART_PINNED pp
            WHERE t.object_name = table_name_in
            AND t.object_type = 'TABLE PARTITION'
            AND t.subobject_name < Part_Maint.partition_name(delete_date)
            AND pp.part_eff_date (+)= Part_Maint.partition_date(t.subobject_name)
            AND pp.part_table_id (+)= table_id_in
            AND pp.part_table_id  IS NULL;

      CURSOR dropqpartition(table_name_in VARCHAR2,
             table_id_in INTEGER, delete_date DATE) IS
            SELECT *
            FROM user_objects t, PART_PINNED pp
            WHERE t.object_name = table_name_in
            AND t.object_type = 'TABLE PARTITION'
            AND t.subobject_name < Part_Maint.partition_name(delete_date, 'L')
            AND pp.part_eff_date (+)= Part_Maint.partition_date(t.subobject_name, 'L')
            AND pp.part_table_id (+)= table_id_in
            AND pp.part_table_id  IS NULL;

      vprc VARCHAR2(30) := 'CleanBehind';

      resource_busy EXCEPTION;
      PRAGMA EXCEPTION_INIT( resource_busy, -00054 );

      needs_attempt BOOLEAN := TRUE;
      attempt_count INTEGER := 0;

   BEGIN

      dbms_application_info.set_action( vprc );
      dbms_application_info.set_client_info( table_desc.table_name );

    IF NOT table_desc.period = 'Q' THEN

      BEGIN

      delete_date := Adm_Date.add_periods(table_desc.task_date, table_desc.period,-1*table_desc.data_retention);
      delete_date := Adm_Date.trunc_to_last_period(delete_date, table_desc.period);

      FOR part_desc IN droppartition(table_desc.table_name, table_desc.part_table_id, delete_date)
      LOOP

        BEGIN
            IF (partition_exists( table_desc.table_name, part_desc.subobject_name ))

            THEN
                sqlstr := 'ALTER TABLE '|| table_desc.table_name ||
                    ' DROP PARTITION '||part_desc.subobject_name;

                needs_attempt := TRUE;

                WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
                   BEGIN
                       Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Dropping Partition '||table_desc.table_name||'.'||part_desc.subobject_name);
                        -- Lock the partition
                        lock_table(table_desc.table_name, part_desc.subobject_name);
                        EXECUTE IMMEDIATE sqlstr;
                        needs_attempt := FALSE;

                   EXCEPTION
                      WHEN resource_busy THEN
                         attempt_count := attempt_count + 1;
                         dbms_application_info.set_client_info( table_desc.table_name ||
                            ': retry ' || TO_CHAR(attempt_count) );
                         dbms_lock.sleep(g_resource_interval);
                   END;
                END LOOP;

            END IF;

        EXCEPTION
            WHEN OTHERS THEN
              IF retval THEN
                retmsg := SQLERRM;
              END IF;
              retval := FALSE;
        END;

      END LOOP;

      END;

    ELSIF table_desc.period = 'Q' THEN

      BEGIN

      delete_date := Adm_Date.add_periods(table_desc.task_date, table_desc.period,-1*table_desc.data_retention);

      delete_date := Adm_Date.trunc_to_last_period(delete_date, table_desc.period);

      FOR part_desc IN dropqpartition(table_desc.table_name, table_desc.part_table_id, delete_date)
      LOOP

        BEGIN
            IF (partition_exists( table_desc.table_name, part_desc.subobject_name ))

            THEN
                sqlstr := 'ALTER TABLE '|| table_desc.table_name ||
                    ' DROP PARTITION '||part_desc.subobject_name;

                needs_attempt := TRUE;

                WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
                   BEGIN
                       Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Dropping Partition '||table_desc.table_name||'.'||part_desc.subobject_name);
                        -- Lock the partition
                        lock_table(table_desc.table_name, part_desc.subobject_name);

                        EXECUTE IMMEDIATE sqlstr;
                        needs_attempt := FALSE;

                   EXCEPTION
                      WHEN resource_busy THEN
                         attempt_count := attempt_count + 1;
                         dbms_application_info.set_client_info( table_desc.table_name ||
                            ': retry ' || TO_CHAR(attempt_count) );
                         dbms_lock.sleep(g_resource_interval);
                   END;
                END LOOP;

            END IF;

        EXCEPTION
            WHEN OTHERS THEN
              IF retval THEN
                retmsg := SQLERRM;
              END IF;
              retval := FALSE;
        END;

      END LOOP;

      END;

    END IF;


      IF ( retval ) THEN
            Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Clean Behind of '||table_desc.table_name||' done.');
      ELSE
            Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Clean Behind of '||table_desc.table_name ||' failed'
                            ||SUBSTR(retmsg,1,200), SQLCODE);
      END IF;

      RETURN retval;

   END cleanbehind;

   -- -----------------------------------------------------------------------
   -- PushAhead - Do ADD PARTITIONS on the high-side of the partitioned table
   -- We probably are creating a bunch of already existing partitions, but
   -- this makes it more fault tolerant.
   -- -----------------------------------------------------------------------
   FUNCTION pushahead( table_desc taskRec ) RETURN BOOLEAN IS
      retval     BOOLEAN := TRUE;
      vReqSpace  INTEGER;
      curdate    DATE;
      partdate   DATE;
      enddate    DATE;
      lastdate   DATE;
      sqlstr     VARCHAR2(4000);
      retmsg     VARCHAR2(9000) := NULL;
      retcode    INTEGER;
      vprc       VARCHAR2(30)   := 'PushAhead';
      overwrite_partition_err EXCEPTION;
      PRAGMA EXCEPTION_INIT(overwrite_partition_err, -14074);
      needs_attempt    BOOLEAN        := TRUE;
      attempt_count    INTEGER        := 0;
      resource_busy    EXCEPTION;
      PRAGMA EXCEPTION_INIT( resource_busy, -00054 );

   BEGIN

      dbms_application_info.set_action( vprc );
      dbms_application_info.set_client_info( table_desc.table_name );

      IF NOT UPPER(table_desc.period) = 'Q' THEN

       BEGIN

      curdate := Adm_Date.trunc_to_last_period(table_desc.task_date, table_desc.period);
      enddate := Adm_Date.add_periods(curdate, table_desc.period, table_desc.future_retention);

      WHILE (curdate <= enddate) LOOP

          partdate := Adm_Date.add_periods(curdate, table_desc.period, 1);

          IF (NOT partition_exists( table_desc.table_name,
                'PART_' || TO_CHAR(curdate, 'YYYYMMDD')))
          THEN

            needs_attempt := TRUE;
            attempt_count := 0;

            WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
               BEGIN

                   sqlstr := 'ALTER TABLE '|| table_desc.table_name ||
                             ' ADD PARTITION PART_' ||
                             TO_CHAR(curdate, 'YYYYMMDD') ||
                             ' VALUES LESS THAN ( to_date(' ||
                             '''' || TO_CHAR(partdate, 'YYYY/MM/DD HH24') || '''' ||
                             ', ' || '''' || 'YYYY/MM/DD HH24' || '''' ||
                             ')) TABLESPACE ' || table_desc.tablespace_name;

                   Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Adding Future Partition '||table_desc.table_name||'.PART_'||TO_CHAR(curdate,'YYYYMMDD'));

                   EXECUTE IMMEDIATE sqlstr;

                   needs_attempt := FALSE;

               EXCEPTION
               WHEN overwrite_partition_err THEN
                    needs_attempt := FALSE; -- This happens all the time...
               WHEN resource_busy THEN
                    attempt_count := attempt_count + 1;
					IF (attempt_count >= g_resource_attempts) THEN
					   retval := FALSE;
                       retmsg := SQLERRM ;
                       retcode := SQLCODE ;
					END IF;   
                    dbms_application_info.set_client_info( table_desc.table_name ||
                       ': retry ' || TO_CHAR(attempt_count) );
                    dbms_lock.sleep(g_resource_interval * attempt_count);
               WHEN OTHERS THEN
                    IF retval THEN
                        retmsg := SQLERRM ;
                        retcode := SQLCODE ;
                    END IF;
                    retval := FALSE;
                    needs_attempt := FALSE;
               END;
            END LOOP;
          END IF;

          curdate := partdate;

      END LOOP;

      IF (retval) THEN
            Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Future Partition Creation for '||table_desc.table_name||' done');
      ELSE
            Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Future Partition Creation for '||table_desc.table_name ||' failed, '
                            ||SUBSTR(retmsg,1,200), retcode);
      END IF;

       RETURN retval;

      END;
      --
      -- For minutely partitions
      --
      ELSIF UPPER(table_desc.period) = 'Q' THEN

        BEGIN

          --
          -- Find date of the last partition
          --
          SELECT MAX(TO_DATE(Partition_Date_Str(partition_name), 'YYYYMMDDHH24MI')) INTO lastdate
          FROM user_tab_partitions
          WHERE table_name = table_desc.table_name;

          curdate := Adm_Date.trunc_to_last_period(lastdate, table_desc.period);
          enddate := Adm_Date.add_periods(SYSDATE, table_desc.period, table_desc.future_retention);

          WHILE (curdate <= enddate) LOOP

            partdate := Adm_Date.add_periods(curdate, table_desc.period, 1);

            IF (NOT partition_exists( table_desc.table_name,
                  'PART_' || TO_CHAR(curdate, 'YYYYMMDDHH24MI')))
            THEN

              needs_attempt := TRUE;
              attempt_count := 0;

              WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
                BEGIN

                  sqlstr := 'ALTER TABLE '|| table_desc.table_name ||
                            ' ADD PARTITION PART_' ||
                            TO_CHAR(curdate, 'YYYYMMDDHH24MI') ||
                            ' VALUES LESS THAN ('||TO_NUMBER(TO_CHAR(partdate, 'YYYYMMDDHH24MI'))||') TABLESPACE ' || table_desc.tablespace_name;

                  Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                           SYSDATE, 'Adding Future Partition '||table_desc.table_name||'.PART_'||TO_CHAR(curdate,'YYYYMMDDHH24MI'));

                  EXECUTE IMMEDIATE sqlstr;

                  needs_attempt := FALSE;

                EXCEPTION
                WHEN overwrite_partition_err THEN
                     needs_attempt := FALSE; -- This happens all the time...
                WHEN resource_busy THEN
                     attempt_count := attempt_count + 1;
					IF (attempt_count >= g_resource_attempts) THEN
					   retval := FALSE;
                       retmsg := SQLERRM ;
                       retcode := SQLCODE ;
					END IF;   

                     dbms_application_info.set_client_info( table_desc.table_name ||
                        ': retry ' || TO_CHAR(attempt_count) );
                     dbms_lock.sleep(g_resource_interval * attempt_count);
                WHEN OTHERS THEN
                     IF retval THEN
                         retmsg := SQLERRM ;
                         retcode := SQLCODE ;
                     END IF;
                     retval := FALSE;
                     needs_attempt := FALSE;
                END;
              END LOOP;
            END IF;

            curdate := partdate;

          END LOOP;

          IF (retval) THEN
                Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Future Partition Creation for '||table_desc.table_name||' done');
          ELSE
                Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Future Partition Creation for '||table_desc.table_name ||' failed, '
                            ||SUBSTR(retmsg,1,200), retcode);
          END IF;

          RETURN retval;
        END;
      END IF;

   END pushahead;

   -- -----------------------------------------------------------------------
   -- PushBehind - Eventually this will do SPLIT PARTITIONs in order to
   -- do the equivalent of PushAhead but on the other end of the partitioned
   -- table
   -- -----------------------------------------------------------------------
   FUNCTION pushbehind( table_desc taskRec ) RETURN BOOLEAN IS
      retval               BOOLEAN := TRUE;
      vReqSpace            INTEGER;
      l_curr_min_date      DATE;
      l_target_min_date    DATE;
      l_curr_align_date    DATE;
      l_past_part_name     VARCHAR2(20);
      l_curr_part_name     VARCHAR2(20);
      l_align_part_name    VARCHAR2(20);
      sqlstr               VARCHAR2(4000);
      needs_attempt        BOOLEAN        := TRUE;
      attempt_count        INTEGER        := 0;
      retmsg               VARCHAR2(9000) := NULL;
      retcode              INTEGER;
      vprc                 VARCHAR2(30)   := 'PushBehind';

      overwrite_partition_err EXCEPTION;
      PRAGMA EXCEPTION_INIT(overwrite_partition_err, -14080);

      resource_busy EXCEPTION;
      PRAGMA EXCEPTION_INIT( resource_busy, -00054 );

   BEGIN

      dbms_application_info.set_action( vprc );
      dbms_application_info.set_client_info( table_desc.table_name );


      IF NOT UPPER(table_desc.period) = 'Q' THEN

      BEGIN

      SELECT MIN(TO_DATE(SUBSTR(partition_name,6,LENGTH(partition_name)), 'YYYYMMDDHH24')) INTO l_curr_min_date
      FROM user_tab_partitions
      WHERE table_name = table_desc.table_name;

      l_target_min_date := Adm_Date.subtract_periods(SYSDATE, table_desc.period, table_desc.past_retention);

      WHILE (l_target_min_date < l_curr_min_date ) LOOP

            l_curr_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_min_date, 'YYYYMMDD')));

            --Set up the period for adding partitions in the past
            IF (table_desc.period='H') THEN
                l_past_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_min_date - 1/24, 'YYYYMMDD')));
            ELSIF (table_desc.period='D') THEN
                l_past_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_min_date - 1, 'YYYYMMDD')));
            ELSIF (table_desc.period='W') THEN
                l_past_part_name  := RTRIM (LTRIM ( 'PART_' || TO_CHAR (TRUNC(l_curr_min_date - 7,'D'), 'YYYYMMDD')));
                l_curr_align_date := TRUNC(l_curr_min_date,'D');
                l_align_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_align_date, 'YYYYMMDD')));
            ELSIF (table_desc.period='M') THEN
                l_past_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (TRUNC(ADD_MONTHS(l_curr_min_date,-1),'MM'), 'YYYYMMDD')));
                l_curr_align_date := TRUNC(l_curr_min_date,'MM');
                l_align_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_align_date, 'YYYYMMDD')));
            END IF;

            IF (partition_exists( table_desc.table_name, l_curr_part_name ))
            THEN

                IF (NOT partition_exists( table_desc.table_name, l_past_part_name ))
                THEN

                   -- Check if partition is aligned with the start of the time period, and if not, re-align
                   IF (l_curr_min_date != l_curr_align_date) THEN

                       l_curr_min_date := l_curr_align_date;

                       sqlstr := 'ALTER TABLE '|| table_desc.table_name
                         || ' split PARTITION '|| l_curr_part_name
                         || ' AT ( TO_DATE('
                         || '''' || TO_CHAR(l_curr_align_date, 'YYYY/MM/DD HH24') || ''''
                         || ', ' || '''' || 'YYYY/MM/DD HH24' || ''''
                         || ')) INTO (PARTITION ' || l_past_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         || ', PARTITION ' || l_align_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         ||')';
                   ELSE
                       sqlstr := 'ALTER TABLE '|| table_desc.table_name
                         || ' split PARTITION '|| l_curr_part_name
                         || ' AT ( TO_DATE('
                         || '''' || TO_CHAR(l_curr_min_date, 'YYYY/MM/DD HH24') || ''''
                         || ', ' || '''' || 'YYYY/MM/DD HH24' || ''''
                         || ')) INTO (PARTITION ' || l_past_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         || ', PARTITION ' || l_curr_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         ||')';
                   END IF;

                   BEGIN

                        needs_attempt := TRUE;

                        WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
                           BEGIN
                             Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                                   SYSDATE, 'Adding Past Partition '||table_desc.table_name||'.'||l_curr_part_name);
                             -- Lock the partition for the curdate_name
                             lock_table(table_desc.table_name, l_curr_part_name);
                             EXECUTE IMMEDIATE sqlstr;

                             needs_attempt := FALSE;

                           EXCEPTION
                              WHEN resource_busy THEN
                                 attempt_count := attempt_count + 1;
                                 dbms_application_info.set_client_info( table_desc.table_name ||
                                    ': retry ' || TO_CHAR(attempt_count) );
                                 dbms_lock.sleep(g_resource_interval);
                           END;
                        END LOOP;

                        -- Only if successful during EXEC
                        IF ( NOT needs_attempt ) THEN

                            IF (partition_exists( table_desc.table_name, l_past_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_past_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                            IF (partition_exists( table_desc.table_name, l_curr_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_curr_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                            IF (partition_exists( table_desc.table_name, l_align_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_align_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                        END IF;

                   EXCEPTION
                      WHEN overwrite_partition_err THEN
                           NULL;        -- This happens all the time...
                      WHEN OTHERS THEN
                            IF retval THEN
                              retmsg := SQLERRM;
                              retcode := SQLCODE;
                            END IF;
                      retval := FALSE;
                   END;

                END IF;

            END IF;

            --Set up the period for adding partitions in the past
            IF (table_desc.period='H') THEN
                l_curr_min_date   := l_curr_min_date - 1/24;
            ELSIF (table_desc.period='D') THEN
                l_curr_min_date := l_curr_min_date - 1;
            ELSIF (table_desc.period='W') THEN
                l_curr_min_date := TRUNC(l_curr_min_date-7,'D');
            ELSIF (table_desc.period='M') THEN
                l_curr_min_date := TRUNC(ADD_MONTHS(l_curr_min_date,-1),'MM');
            END IF;

         END LOOP;

       END;

      -- For Q partitions
      ELSIF UPPER(table_desc.period) = 'Q' THEN
      BEGIN

      SELECT MIN(TO_DATE(Partition_Date_Str(partition_name), 'YYYYMMDDHH24MI')) INTO l_curr_min_date
      FROM user_tab_partitions
      WHERE table_name = table_desc.table_name;

      l_target_min_date := Adm_Date.subtract_periods(SYSDATE, table_desc.period, table_desc.past_retention);

      WHILE (l_target_min_date < l_curr_min_date ) LOOP

            l_curr_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_min_date, 'YYYYMMDDHH24MI')));

            --Set up the period for adding partitions in the past

             l_past_part_name := RTRIM (LTRIM ( 'PART_' || TO_CHAR (l_curr_min_date - 1/96, 'YYYYMMDDHH24MI')));

            IF (partition_exists( table_desc.table_name, l_curr_part_name ))
            THEN

                IF (NOT partition_exists( table_desc.table_name, l_past_part_name ))
                THEN

                   -- Check if partition is aligned with the start of the time period, and if not, re-align
                   IF (l_curr_min_date != l_curr_align_date) THEN

                       l_curr_min_date := l_curr_align_date;

                       sqlstr := 'ALTER TABLE '|| table_desc.table_name
                         || ' split PARTITION '|| l_curr_part_name
                         || ' AT ( '|| TO_NUMBER(TO_CHAR(l_curr_align_date, 'YYYYMMDDHH24MI'))
                         || ') INTO (PARTITION ' || l_past_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         || ', PARTITION ' || l_align_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         ||')';
                   ELSE
                       sqlstr := 'ALTER TABLE '|| table_desc.table_name
                         || ' split PARTITION '|| l_curr_part_name
                         || ' AT ('|| TO_NUMBER(TO_CHAR(l_curr_min_date, 'YYYYMMDDHH24MI'))
                         || ') INTO (PARTITION ' || l_past_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         || ', PARTITION ' || l_curr_part_name
                         || ' TABLESPACE ' || table_desc.tablespace_name
                         ||')';
                   END IF;


                   BEGIN

                        needs_attempt := TRUE;

                        WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP
                           BEGIN

                             Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                                   SYSDATE, 'Adding Past Partition '||table_desc.table_name||'.'||l_curr_part_name);
                             -- Lock the partition for the curdate_name
                             lock_table(table_desc.table_name, l_curr_part_name);

                             EXECUTE IMMEDIATE sqlstr;

                             needs_attempt := FALSE;

                           EXCEPTION
                              WHEN resource_busy THEN
                                 attempt_count := attempt_count + 1;
                                 dbms_application_info.set_client_info( table_desc.table_name ||
                                    ': retry ' || TO_CHAR(attempt_count) );
                                 dbms_lock.sleep(g_resource_interval);
                           END;
                        END LOOP;

                        -- Only if successful during EXEC
                        IF ( NOT needs_attempt ) THEN

                            IF (partition_exists( table_desc.table_name, l_past_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_past_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                            IF (partition_exists( table_desc.table_name, l_curr_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_curr_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                            IF (partition_exists( table_desc.table_name, l_align_part_name ))
                            THEN
                                EXECUTE IMMEDIATE 'ALTER TABLE '||table_desc.table_name ||' MODIFY PARTITION '
                                 || l_align_part_name
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
                            END IF;

                        END IF;

                   EXCEPTION
                      WHEN overwrite_partition_err THEN
                           NULL;        -- This happens all the time...
                      WHEN OTHERS THEN
                            IF retval THEN
                              retmsg := SQLERRM;
                              retcode := SQLCODE;
                            END IF;
                      retval := FALSE;
                   END;

                END IF;

            END IF;

            --Set up the period for adding partitions in the past
                l_curr_min_date   := l_curr_min_date - 1/96;

         END LOOP;

      EXCEPTION
      WHEN NO_DATA_FOUND THEN
         NULL;
      END;

      END IF;

      IF (retval) THEN
            Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Past Partition Creation for '||table_desc.table_name||' done.');
      ELSE
            Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Past Partition Creation for '||table_desc.table_name ||' failed. '
                            ||SUBSTR(retmsg,1,200), retcode);
      END IF;

      RETURN retval;

   END pushbehind;

   -- -----------------------------------------------------------------------
   -- MovePartitions - Because of the way Oracle splits full blocks, *most*
   -- IOTs end up with ~50% block utilization.  We can go in an do a MOVE
   -- PARTITION and coalesce all of the blocks to be 90%+ block utilization.
   -- This improves data density (more real data per block) and just might
   -- cause one or more extents to be unnecessary, thus saving us some disk
   -- space.
   -- -----------------------------------------------------------------------
   FUNCTION movepartitions( table_desc taskRec ) RETURN BOOLEAN IS

      CURSOR valid_iots(table_name_in VARCHAR2) IS
            SELECT table_name
            FROM user_tables
            WHERE table_name = table_name_in
            AND iot_type = 'IOT';

      sqlstr         VARCHAR2(4000);
      move_part      VARCHAR2(50);
      retval         BOOLEAN        := TRUE;
      retcode        INTEGER;
      retmsg         VARCHAR2(9000) := NULL;
      needs_attempt  BOOLEAN        := TRUE;
      attempt_count  INTEGER        := 0;
      vprc           VARCHAR2(30)   := 'MovePartitions';

      -- We need a local to keep the Move and Analyze separate
      l_task_id INTEGER;

      resource_busy EXCEPTION;
      PRAGMA EXCEPTION_INIT( resource_busy, -00054 );

   BEGIN


      /* MovePartitions should only be ran for IOT tables */
      FOR valid_tabs IN valid_iots(table_desc.table_name)
      LOOP

          dbms_application_info.set_action( vprc );
          dbms_application_info.set_client_info( table_desc.table_name || '.' || move_part );

          IF NOT UPPER(table_desc.period) = 'Q' THEN
            move_part := 'PART_' || TO_CHAR(table_desc.task_date, 'YYYYMMDD');
          ELSIF UPPER(table_desc.period) = 'Q' THEN
            move_part := 'PART_' || TO_CHAR(table_desc.task_date, 'YYYYMMDDHH24MI');
          END IF;

          l_task_id := generic_start_task(table_desc.job_id, 'PART_MAINT: Starting Move of '|| table_desc.table_name || '.' || move_part);

          sqlstr := 'ALTER TABLE '|| table_desc.table_name ||
                    ' MOVE PARTITION ' || move_part ||
                    ' TABLESPACE ' || table_desc.tablespace_name ||
                    ' STORAGE ( INITIAL ' || getextentsize(table_desc.tablespace_name) ||
                    ' NEXT ' || getextentsize(table_desc.tablespace_name) ||')';

          IF (partition_exists(table_desc.table_name, move_part)) THEN

              BEGIN

                  needs_attempt := TRUE;

                  WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) ) LOOP

                       BEGIN
                            Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                                   SYSDATE, 'Moving Partition '||table_desc.table_name||'.'||move_part);
                            -- Lock the partition
                            lock_table(table_desc.table_name, move_part);
                            EXECUTE IMMEDIATE sqlstr;

                            needs_attempt := FALSE;

                       EXCEPTION
                          WHEN resource_busy THEN
                             attempt_count := attempt_count + 1;
                             dbms_application_info.set_client_info( table_desc.table_name
                                || '.'
                                || move_part
                                || ': retry '
                                || TO_CHAR(attempt_count) );
                             dbms_lock.sleep(g_resource_interval);
                       END;

                  END LOOP;

              EXCEPTION
              WHEN OTHERS THEN
                  IF retval THEN
                     retmsg := SQLERRM;
                     retcode := SQLCODE;
                  END IF;
                  retval := FALSE;
              END;
          END IF;

          IF (retval) THEN
              Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Move of '||table_desc.table_name||' done.');
          ELSE
              Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Move of '||table_desc.table_name ||' failed. '
                            ||SUBSTR(retmsg,1,200), retcode);
          END IF;

      END LOOP;

      RETURN retval;

   END movepartitions;

   -- -----------------------------------------------------------------------
   -- AnalyzePartitions - Analyze the paritions so we can use the CBO.  Do
   -- the analyze *after* the move.  The statistics include such things as
   -- number of blocks and average number of blocks per key, and these will
   -- change on the MOVE PARTITION.
   -- -----------------------------------------------------------------------
   FUNCTION analyzepartitions( table_desc taskRec ) RETURN BOOLEAN IS

      retval          BOOLEAN := TRUE;
      sqlstr          VARCHAR2(4000);
      analyze_part    VARCHAR2(30);
      l_check         NUMBER        := NULL;
      parallel_degree NUMBER        := NULL;
      retcode         INTEGER;
      retmsg          VARCHAR2(200) := NULL;
      vprc            VARCHAR2(30)  := 'AnalyzePartitions';

      -- We need a local to keep the Move and Analyze separate
      l_task_id INTEGER;
   BEGIN

      -- Selectthe analyze_mask against the current day.
      -- Check to see if analyze should be ran on current day.
      l_check := TO_NUMBER(SUBSTR(table_desc.analyze_mask,TO_CHAR(SYSDATE,'D'),1));

      IF ( l_check = 1 ) THEN

          IF NOT UPPER(table_desc.period) = 'Q' THEN
            analyze_part := 'PART_' || TO_CHAR(table_desc.task_date, 'YYYYMMDD');
          ELSIF UPPER(table_desc.period) = 'Q' THEN
            analyze_part := 'PART_' || TO_CHAR(table_desc.task_date, 'YYYYMMDDHH24MI');
          END IF;

          dbms_application_info.set_action( vprc );
          dbms_application_info.set_client_info( table_desc.table_name || '.' || analyze_part );

          l_task_id := generic_start_task(table_desc.job_id, 'PART_MAINT: Starting Analyze of '|| table_desc.table_name || '.' || analyze_part);

          IF (partition_exists(table_desc.table_name, analyze_part)) THEN
             BEGIN
                IF(g_stats_parallel > 0)THEN
                  parallel_degree := g_stats_parallel;
                END IF;

                Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                                   SYSDATE, 'Analyzing Partition '||table_desc.table_name||'.'||analyze_part);
                DBMS_STATS.GATHER_TABLE_STATS(
                  ownname          => USER,
                  tabname          => table_desc.table_name,
                  partname         => analyze_part,
                  estimate_percent => g_analyze_pct,
                  DEGREE           => parallel_degree,
                  method_opt       => 'FOR ALL COLUMNS SIZE 1',
                  granularity      => 'PARTITION',
                  block_sample     => FALSE,
                  CASCADE          => FALSE);


             EXCEPTION
                WHEN OTHERS THEN
                      IF retval THEN
                        retmsg := SQLERRM;
                        retcode := SQLCODE;
                      END IF;
                      retval := FALSE;
             END;
          END IF;


          IF (retval) THEN
              Log_Pkg.log_info_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Analyze of '||table_desc.table_name||' done.');
          ELSE
              Log_Pkg.log_warn_table(g_module_name, vprc, table_desc.job_id, table_desc.task_id,
                            SYSDATE, 'Analyze of '||table_desc.table_name ||' failed. '
                            ||SUBSTR(retmsg,1,200), retcode);
          END IF;

      END IF;

      RETURN retval;

   END analyzepartitions;

   -- -----------------------------------------------------------------------
   -- LockTable - obtain an exclusive lock on the table.  To be called before each DDL.
   --           - enhanced to use partition name, if present.
   -- -----------------------------------------------------------------------
   PROCEDURE lock_table(table_name VARCHAR2, partition_name VARCHAR2 DEFAULT NULL) IS
      sqlstr VARCHAR2(200);
   BEGIN

      IF ( partition_name IS NULL ) THEN
         sqlstr := 'LOCK TABLE '
                   || table_name
                   || ' IN EXCLUSIVE MODE';
      ELSE
         -- Add the partition name
         sqlstr := 'LOCK TABLE '
                   || table_name
                   || ' PARTITION ( ' || partition_name || ' ) '
                   || ' IN EXCLUSIVE MODE';
      END IF;

      EXECUTE IMMEDIATE sqlstr;

   END lock_table;

   -- -----------------------------------------------------------------------
   -- flush_schema_monitoring_info: procedure used to flush the records of
   -- modifications to the tables into USER_TAB_MODIFICATIONS
   -- -----------------------------------------------------------------------
   PROCEDURE flush_schema_monitoring_info AS
   BEGIN
       EXECUTE IMMEDIATE 'BEGIN DBMS_STATS.FLUSH_DATABASE_MONITORING_INFO; END;';
   END flush_schema_monitoring_info;

   -- -----------------------------------------------------------------------
   -- get_tunable: function to get tunable parameters from system variables
   -- or set to defaults.
   -- -----------------------------------------------------------------------
   FUNCTION get_tunable( param_name VARCHAR2, param_default INTEGER) RETURN INTEGER IS
     parameter_val INTEGER;
   BEGIN

      SELECT TRUNC( TO_NUMBER( VALUE ) ) INTO parameter_val
      FROM ADM_SYSTEM_VARIABLES
      WHERE UPPER(NAME) = UPPER( param_name );

      RETURN NVL(parameter_val, param_default);

   EXCEPTION
     WHEN OTHERS THEN
       RETURN param_default;
   END get_tunable;

   -- -----------------------------------------------------------------------
   -- init_tunable_parameters:procedure to initalise the tunable parameters
   -- from system variables or set to defaults.
   -- -----------------------------------------------------------------------
   PROCEDURE init_tunable_parameters IS
    vprc VARCHAR2(30) := 'Init_Tunable_Parameters';
   BEGIN
     g_max_run_time    := get_tunable('PartMaintMaxRunTime', g_max_run_time);
     g_analyze_pct     := get_tunable('PartMaintAnalyzePct', g_analyze_pct);
     g_update_move_pct := get_tunable('PartMaintUpdatePct', g_update_move_pct);
     g_stats_parallel  := get_tunable('PartMaintStatsParallel', g_stats_parallel);
     g_poll_interval   := get_tunable( 'PartMaintPollInterval', g_poll_interval );
     g_retry_interval  := get_tunable( 'PartMaintRetryInterval', g_retry_interval );
     -- Parameters for part_tasks, we always needs a least g_min_part_sessions
     g_part_sessions   := GREATEST( get_tunable( 'PartMaintSessions', g_part_sessions ), g_min_part_sessions );
   END init_tunable_parameters;

   -- -----------------------------------------------------------------------
   -- job_interrupted - Check and see if this job has been marked as killed,
   -- interrupted, done, etc.
   -- -----------------------------------------------------------------------
   FUNCTION job_interrupted(job_id_in IN INTEGER) RETURN BOOLEAN IS

      CURSOR check_status_cur IS
         SELECT DISTINCT status
         FROM v$session
         WHERE audsid = USERENV('SESSIONID')
         AND module='PartMaint';

      ret_val         BOOLEAN := FALSE;
      start_time      DATE;
      taskid          INTEGER      := NULL;
      vprc            VARCHAR2(30) := 'Job_Interrupted';
      session_status sys.v_$session.Status%TYPE;

   BEGIN

      FOR l_status IN check_status_cur
      LOOP

          BEGIN

             IF ( l_status.status = 'KILLED' ) THEN
                taskid := generic_start_task(job_id_in, g_module_name||', Job Interrupted: Session was KILLED.');
                generic_end_task( job_id_in, taskid, g_module_name||', Job Interrupted: Session was KILLED.');
                Log_Pkg.log_warn_table(g_module_name, vprc, job_id_in, NULL,
                        SYSDATE, 'Job Interrupted: Session was KILLED', NULL);
                ret_val := TRUE;

             ELSIF ( ( SYSDATE - start_time ) *60*60*24 > g_max_run_time ) THEN
                taskid := generic_start_task(job_id_in, g_module_name||', Job Interrupted: Exceeded maximum run time.');
                generic_end_task( job_id_in, taskid, g_module_name||', Job Interrupted: Exceeded maximum run time.');
                Log_Pkg.log_warn_table(g_module_name, vprc, job_id_in, NULL,
                        SYSDATE, 'Job Interrupted: Exceeded maximum run time.', NULL);
                ret_val := TRUE;

             END IF;

             ret_val := FALSE;

          EXCEPTION
          WHEN OTHERS THEN
            Log_Pkg.log_warn_table(g_module_name, vprc, job_id_in, NVL(taskid,NULL),
                        SYSDATE, 'Job Interrupted exception.'||SQLERRM, SQLCODE);
            DBMS_OUTPUT.PUT_LINE('Job Interrupted exception.'||SQLERRM);
            ret_val := FALSE;
          END;

      END LOOP;

      RETURN ret_val;

   END job_interrupted;

   -- -----------------------------------------------------------------------
   -- Create the Add/Drop operations
   -- -----------------------------------------------------------------------
   PROCEDURE create_ad_ops( p_period IN CHAR,
                             p_eff_date IN DATE,
                             p_jobid IN INTEGER )
   IS

      p_task taskRec;

      CURSOR alltables(period_in CHAR) IS
        SELECT *
          FROM PART_TABLES
          WHERE period = period_in
          ORDER BY tablespace_name, table_name;

   BEGIN

      -- For Loop to perform the necessary Add/Drop/Split operations where needed
      FOR table_desc IN alltables(p_period) LOOP

         -- Load the record
         p_task.part_table_id    := table_desc.part_table_id;
         p_task.table_name       := table_desc.table_name;
         p_task.tablespace_name  := table_desc.tablespace_name;
         p_task.period           := table_desc.period;
         p_task.data_retention   := table_desc.data_retention;
         p_task.past_retention   := table_desc.past_retention;
         p_task.future_retention := table_desc.future_retention;
         p_task.needs_move       := table_desc.needs_move;
         p_task.needs_export     := table_desc.needs_export;
         p_task.needs_purge      := table_desc.needs_purge;
         p_task.needs_analyze    := table_desc.needs_analyze;
         p_task.analyze_mask     := table_desc.analyze_mask;

         IF (table_desc.future_retention IS NOT NULL) THEN
            create_add_task ( p_jobid, p_eff_date, p_task );
         END IF;

         IF (table_desc.past_retention IS NOT NULL ) THEN
            create_split_task ( p_jobid, p_eff_date, p_task );
         END IF;

         IF (table_desc.data_retention IS NOT NULL AND table_desc.needs_purge = 'Y') THEN
            create_drop_task ( p_jobid, p_eff_date, p_task );
         END IF;

      END LOOP;

   EXCEPTION
      WHEN OTHERS THEN
         RAISE;

   END create_ad_ops;

   -- -----------------------------------------------------------------------
   -- Create the Split operations
   -- -----------------------------------------------------------------------
   PROCEDURE create_split_ops( p_period IN CHAR,
                             p_eff_date IN DATE,
                             p_jobid IN INTEGER )
   IS

      p_task taskRec;

      CURSOR alltables(period_in CHAR) IS
        SELECT *
          FROM PART_TABLES
          WHERE period = period_in
          ORDER BY tablespace_name, table_name;

   BEGIN

      -- For Loop to perform the necessary Add/Drop/Split operations where needed
      FOR table_desc IN alltables(p_period) LOOP

         -- Load the record
         p_task.part_table_id    := table_desc.part_table_id;
         p_task.table_name       := table_desc.table_name;
         p_task.tablespace_name  := table_desc.tablespace_name;
         p_task.period           := table_desc.period;
         p_task.data_retention   := table_desc.data_retention;
         p_task.past_retention   := table_desc.past_retention;
         p_task.future_retention := table_desc.future_retention;
         p_task.needs_move       := table_desc.needs_move;
         p_task.needs_export     := table_desc.needs_export;
         p_task.needs_purge      := table_desc.needs_purge;
         p_task.needs_analyze    := table_desc.needs_analyze;
         p_task.analyze_mask     := table_desc.analyze_mask;


         IF (table_desc.data_retention IS NOT NULL) THEN
            create_split_task ( p_jobid, p_eff_date, p_task );
         END IF;

      END LOOP;

   EXCEPTION
      WHEN OTHERS THEN
         RAISE;

   END create_split_ops;

   -- -----------------------------------------------------------------------
   -- Create the Move/Analyze operations
   -- -----------------------------------------------------------------------
   PROCEDURE create_ma_ops(startdate IN DATE,
                           enddate IN DATE,
                           period_in IN CHAR,
                           usetabmodifications IN BOOLEAN DEFAULT FALSE,
                           jobidparam IN INTEGER DEFAULT NULL)
   IS
      p_task taskRec;

      part_date     DATE;
      eff_startdate DATE := TRUNC(startdate);
      eff_enddate   DATE := TRUNC(enddate);

      -- All tables
      CURSOR alltables(startdate_in DATE, enddate_in DATE, period_in CHAR) IS
        SELECT
          p.part_table_id,
          p.table_name,
          p.tablespace_name,
          p.period,
          p.data_retention,
          p.past_retention,
          p.future_retention,
          p.needs_move,
          p.needs_export,
          p.needs_purge,
          p.needs_analyze,
          p.analyze_mask,
          uo.subobject_name part_name
        FROM PART_TABLES p, user_objects uo
        WHERE p.period = period_in
          AND p.table_name = uo.object_name
          AND uo.object_type = 'TABLE PARTITION'
          AND (p.needs_move = 'Y' OR p.needs_analyze = 'Y')
          AND  uo.subobject_name
          BETWEEN Part_Maint.partition_name(startdate_in)
          AND Part_Maint.partition_name(enddate_in)
          AND uo.subobject_name != Part_Maint.partition_name(TRUNC(SYSDATE))
          ORDER BY uo.subobject_name, p.table_name;

      -- Modified tables
      CURSOR modtables(startdate_in DATE, enddate_in DATE, period_in CHAR, update_pct_in INTEGER) IS
        SELECT p.part_table_id,
           p.table_name,
           p.tablespace_name,
           p.period,
           p.data_retention,
           p.past_retention,
           p.future_retention,
           p.needs_move,
           p.needs_export,
           p.needs_purge,
           p.needs_analyze,
           p.analyze_mask,
           utm.partition_name part_name
        FROM PART_TABLES p, user_tab_modifications utm, user_tab_partitions utp
        WHERE p.period = period_in
        AND p.table_name = utm.table_name
        AND (p.needs_move = 'Y' OR p.needs_analyze = 'Y')
        AND  utm.partition_name
          BETWEEN Part_Maint.partition_name(startdate_in)
          AND Part_Maint.partition_name(enddate_in)
        AND utm.partition_name != Part_Maint.partition_name(TRUNC(SYSDATE))
        AND utm.table_name = utp.table_name
        AND utm.partition_name = utp.partition_name
        AND ( utp.last_analyzed IS NULL
          OR utm.inserts > 0
          OR utm.updates > (NVL(utp.num_rows,0)*update_pct_in)/100
          OR utm.deletes > 0
          OR utm.truncated = 'YES')
        ORDER BY utm.inserts, utm.partition_name, p.table_name;
   BEGIN

      IF (startdate IS NULL) THEN
         eff_startdate := TO_DATE('19700101', 'YYYYMMDD');
      END IF;

      IF (eff_enddate IS NULL) THEN
         eff_enddate := TO_DATE('20380101', 'YYYYMMDD');
      END IF;

      -- Flust Monitoring info before MA tasks are created
      flush_schema_monitoring_info();

      -- Only create OPs where needed
      IF (usetabmodifications) THEN
           FOR table_desc IN
           modtables(eff_startdate, eff_enddate, period_in, g_update_move_pct) LOOP

                -- Load the record
                p_task.part_table_id    := table_desc.part_table_id;
                p_task.table_name       := table_desc.table_name;
                p_task.tablespace_name  := table_desc.tablespace_name;
                p_task.period           := table_desc.period;
                p_task.data_retention   := table_desc.data_retention;
                p_task.past_retention   := table_desc.past_retention;
                p_task.future_retention := table_desc.future_retention;
                p_task.needs_move       := table_desc.needs_move;
                p_task.needs_export     := table_desc.needs_export;
                p_task.needs_purge      := table_desc.needs_purge;
                p_task.needs_analyze    := table_desc.needs_analyze;
                p_task.analyze_mask     := table_desc.analyze_mask;
                p_task.part_name        := table_desc.part_name;

                part_date := Part_Maint.partition_date(table_desc.part_name);

               -- The M+A are combined due to Oracle issues
               IF (table_desc.needs_move = 'Y' OR table_desc.needs_analyze = 'Y') THEN
                  create_ma_task( jobidparam, part_date, p_task );
               END IF;

           END LOOP;
      ELSE
           FOR table_desc IN
               alltables(eff_startdate, NVL(eff_enddate, eff_startdate), period_in) LOOP

                   -- Load the record
                   p_task.part_table_id    := table_desc.part_table_id;
                   p_task.table_name       := table_desc.table_name;
                   p_task.tablespace_name  := table_desc.tablespace_name;
                   p_task.period           := table_desc.period;
                   p_task.data_retention   := table_desc.data_retention;
                   p_task.past_retention   := table_desc.past_retention;
                   p_task.future_retention := table_desc.future_retention;
                   p_task.needs_move       := table_desc.needs_move;
                   p_task.needs_export     := table_desc.needs_export;
                   p_task.needs_purge      := table_desc.needs_purge;
                   p_task.needs_analyze    := table_desc.needs_analyze;
                   p_task.analyze_mask     := table_desc.analyze_mask;
                   p_task.part_name        := table_desc.part_name;

                   part_date := Part_Maint.partition_date(table_desc.part_name);

                   -- The M+A are combined dues to Oracle issues
                   IF (table_desc.needs_move = 'Y' OR table_desc.needs_analyze = 'Y') THEN
                      create_ma_task( jobidparam, part_date, p_task );
                   END IF;

         END LOOP;

      END IF;

   EXCEPTION
      WHEN OTHERS THEN
         RAISE;

   END create_ma_ops;

   -- -----------------------------------------------------------------------
   -- Create the part_task record including the OP info.
   --     This is a child record of the active_task record.
   -- -----------------------------------------------------------------------
   PROCEDURE add_part_task( p_operation IN INTEGER,
                                p_jobid     IN INTEGER,
                                p_task_id   IN INTEGER,
                                p_date      IN DATE,
                                p_task      IN taskRec )
   IS
   BEGIN

      INSERT
         INTO PART_TASK (
         job_id,
         task_id,
         status_code,
         operation,
         task_date,
         part_table_id,
         table_name,
         tablespace_name,
         period,
         data_retention,
         past_retention,
         future_retention,
         needs_move,
         needs_export,
         needs_purge,
         needs_analyze,
         analyze_mask,
         part_name
         )
      VALUES (
         p_jobid,
         p_task_id,
         g_OPS_READY,                   -- Task is READY
         p_operation,                   -- operation needed
         p_date,                        -- Effective date or part_date, depends on task
         p_task.part_table_id,
         p_task.table_name,
         p_task.tablespace_name,
         p_task.period,
         p_task.data_retention,
         p_task.past_retention,
         p_task.future_retention,
         p_task.needs_move,
         p_task.needs_export,
         p_task.needs_purge,
         p_task.needs_analyze,
         p_task.analyze_mask,
         p_task.part_name);

   END add_part_task;

   -- -----------------------------------------------------------------------
   -- Create an ADD task
   -- -----------------------------------------------------------------------
   PROCEDURE create_add_task( p_jobid IN INTEGER,
                              p_eff_date IN DATE,
                              p_task IN taskRec )
   IS
      task_id INTEGER;
   BEGIN
      task_id := generic_start_task(p_jobid, g_module_name||', ADD: ' || p_task.table_name);
      add_part_task( g_Add_Op, p_jobid, task_id, p_eff_date, p_task );
   END create_add_task;

   -- -----------------------------------------------------------------------
   -- Create a DROP task
   -- -----------------------------------------------------------------------
   PROCEDURE create_drop_task( p_jobid IN INTEGER,
                               p_eff_date IN DATE,
                               p_task IN taskRec )
   IS
      task_id INTEGER;
   BEGIN
      task_id := generic_start_task(p_jobid, g_module_name||', DROP: ' || p_task.table_name);
      add_part_task( g_Drop_Op, p_jobid, task_id, p_eff_date, p_task );
   END create_drop_task;

   -- -----------------------------------------------------------------------
   -- Create a SPLIT task
   -- -----------------------------------------------------------------------
   PROCEDURE create_split_task( p_jobid IN INTEGER,
                                p_eff_date IN DATE,
                                p_task IN taskRec )
   IS
      task_id INTEGER;
   BEGIN
      task_id := generic_start_task(p_jobid, g_module_name||', SPLIT: ' || p_task.table_name);
      add_part_task( g_Split_Op, p_jobid, task_id, p_eff_date, p_task );
   END create_split_task;

   -- -----------------------------------------------------------------------
   -- Create a Move Analyze task
   -- Composite task for Move and Analyze to work around some Oracle
   -- issues with row cache enqueues. The Move and Analyze in separate
   -- sessions appears to block each other.
   -- -----------------------------------------------------------------------
   PROCEDURE create_ma_task( p_jobid IN INTEGER,
                             p_eff_date IN DATE,
                             p_task IN taskRec )
   IS
      task_id INTEGER;
   BEGIN
      task_id := generic_start_task(p_jobid, g_module_name||', MOVE + ANALYZE: '
                    || p_task.table_name|| '.'|| p_task.part_name||', mask:'||p_task.analyze_mask);
      add_part_task( g_MoveAnalyze_Op, p_jobid, task_id, p_eff_date, p_task );
   END create_ma_task;

   -- -----------------------------------------------------------------------
   -- Create the sessions for this run of part_maint
   -- -----------------------------------------------------------------------
   PROCEDURE create_part_sessions( p_jobid IN INTEGER, p_part_sessions IN INTEGER )
   IS
      part_task_id INTEGER;
      task_id INTEGER;
      vprc    VARCHAR2(30) := 'Create_Part_Session';
   BEGIN

      FOR i IN 1..p_part_sessions LOOP

         -- Create a task for the session so that we can track the status
         task_id := generic_start_task( p_jobid, g_module_name||', Session: preparing to submit' );

         -- Create session using DBMS_JOB.SUBMIT. Pass job_id and task_id to allow
         -- the job the ability to update its own task
         DBMS_JOB.SUBMIT( part_task_id, 'part_maint.part_session( JOB, '
                     || TO_CHAR(p_jobid) || ', ' || TO_CHAR(task_id) || ' );' );

         -- Update the global table for session ids
         session_id_tab.EXTEND;
         session_id_tab( session_id_tab.LAST ) := session_id_rowtype( part_task_id );

         Log_Pkg.log_info_table(g_module_name, vprc, p_jobid, task_id,  SYSDATE, 'Submitting session '||part_task_id);

      END LOOP;

      EXCEPTION
         WHEN OTHERS THEN
            Log_Pkg.log_warn_table(g_module_name, vprc, p_jobid, task_id,  SYSDATE,
                    'Exception Submitting session '||SQLERRM, SQLCODE);
            RAISE;

   END create_part_sessions;

   -- -----------------------------------------------------------------------
   -- Update all READY tasks to ABORT. We are trying to starve
   -- the get_next_task routine so that the sessions will complete.
   -- -----------------------------------------------------------------------
   PROCEDURE abort_tasks( p_jobid IN INTEGER )
   IS
      l_message VARCHAR2(200);
      b_process_it BOOLEAN:=TRUE;

      CURSOR c_get_task  ( p_jobid INTEGER)IS
          SELECT * FROM PART_TASK
          WHERE job_id = p_jobid
          AND status_code = g_OPS_READY
          FOR UPDATE OF status_code;

      vprc VARCHAR2(30) := 'Abort_Tasks';
   BEGIN

      -- Let them know we are starting
      dbms_application_info.set_module( g_module_name, vprc );

      Lock_Util.get_x(g_task_lock_name);

      FOR l_task IN c_get_task ( p_jobid) LOOP
        b_process_it:=TRUE;

         -- Format the message
         CASE l_task.operation
            WHEN g_Drop_Op THEN
                l_message:=g_module_name||', Cleanup of '||l_task.table_name;
            WHEN g_Add_Op THEN
                 l_message:=g_module_name||', Forward roll-out of '||l_task.table_name;
            WHEN g_Split_Op THEN
              l_message:=g_module_name||', Backward roll-out of '||l_task.table_name;
            WHEN g_MoveAnalyze_Op THEN
              l_message:=g_module_name||', Move + Analyze: '|| l_task.table_name|| '.'|| l_task.part_name;
            ELSE
              -- Unknown task, ignore it. Should be handled by the process that created it
              b_process_it:=FALSE;
         END CASE;

         IF ( b_process_it) THEN
            -- Update the dangling task_status
            -- ensure that the message is not over the limit sze
            l_message:=SUBSTR(l_message,1,400) || ' Skipped';
            Log_Pkg.log_info_table(g_module_name, vprc, p_jobid, NULL,
                            SYSDATE, l_message);

            -- remove the task
            UPDATE PART_TASK
            SET status_code = g_OPS_ABORT
            WHERE CURRENT OF c_get_task;
         END IF;

      END LOOP;

      Lock_Util.release(g_task_lock_name);
      COMMIT;

   EXCEPTION
      WHEN OTHERS THEN
         Lock_Util.release(g_task_lock_name);
         Log_Pkg.log_warn_table(g_module_name, vprc, p_jobid, NULL,
                            SYSDATE, SQLERRM, SQLCODE);
         RAISE;

   END abort_tasks;

   -- -----------------------------------------------------------------------
   -- Have all of the sessions completed?
   -- -----------------------------------------------------------------------
   FUNCTION all_sessions_complete RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      -- Use our global session id table
      CURSOR job_cur IS
         SELECT 1
         FROM user_jobs uj, TABLE( CAST( session_id_tab AS session_id_tabletype ) ) s
         WHERE s.job_id = uj.job;

      job_rec job_cur%ROWTYPE;

   BEGIN

      -- We are looking to see if anything is still processing
      OPEN job_cur;
      FETCH job_cur INTO job_rec;
      l_status := job_cur%NOTFOUND;
      CLOSE job_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( job_cur%ISOPEN ) THEN
            CLOSE job_cur;
         END IF;

         RAISE;

   END all_sessions_complete;

   -- -----------------------------------------------------------------------
   -- Have all of the tasks completed with a SUCCESS?
   -- -----------------------------------------------------------------------
   FUNCTION all_tasks_success( p_jobid IN INTEGER ) RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT 1
         FROM PART_TASK
         WHERE job_id = c_job_id
         AND status_code != c_code;

      task_rec task_cur%ROWTYPE;

   BEGIN

      -- We are looking to see if any tasks are NOT marked a SUCCESS
      OPEN task_cur( p_jobid, g_OPS_SUCCESS );
      FETCH task_cur INTO task_rec;
      l_status := task_cur%NOTFOUND;
      CLOSE task_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( task_cur%ISOPEN ) THEN
            CLOSE task_cur;
         END IF;

         RAISE;

   END all_tasks_success;

   -- -----------------------------------------------------------------------
   -- Are there READY tasks waiting?
   -- -----------------------------------------------------------------------
   FUNCTION waiting_tasks( p_jobid IN INTEGER ) RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT 1
         FROM PART_TASK
         WHERE job_id = c_job_id
         AND status_code = c_code;

      task_rec task_cur%ROWTYPE;
   BEGIN

      -- We are looking to see if any tasks are marked as READY
      OPEN task_cur( p_jobid, g_OPS_READY );
      FETCH task_cur INTO task_rec;
      l_status := task_cur%FOUND;
      CLOSE task_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( task_cur%ISOPEN ) THEN
            CLOSE task_cur;
         END IF;

         RAISE;

   END waiting_tasks;


   -- -----------------------------------------------------------------------
   -- This is the SESSION for part_maint.
   -- Once started it continues to run as long as it receives new tasks.
   -- -----------------------------------------------------------------------
   PROCEDURE part_session( p_orajobid IN INTEGER, p_jobid IN INTEGER, p_taskid IN INTEGER)
   IS
      success BOOLEAN := TRUE;
      needs_task BOOLEAN := TRUE;
      l_task taskRec;

      vprc VARCHAR2(30) := 'part_session';
   BEGIN

      -- Let them know we are starting
      dbms_application_info.set_module( g_module_name, vprc );

      -- init the session
      init_tunable_parameters();

      WHILE ( needs_task ) LOOP

         -- Get the next task
         get_next_task(p_jobid, l_task);

         IF (l_task.task_id IS NULL ) THEN
            -- No more work available
            needs_task := FALSE;
         ELSE

            -- Do the task
            CASE l_task.operation
               WHEN g_Drop_Op THEN
                  -- Drop
                  success := cleanbehind( l_task );
               WHEN g_Add_Op THEN
                  -- Add
                  success := pushahead( l_task );
               WHEN g_Split_Op THEN
                  -- Split
                  success := pushbehind( l_task );
               WHEN g_MoveAnalyze_Op THEN
                  IF ( l_task.needs_move = 'Y' ) THEN
                     -- Move
                     success := movepartitions( l_task );
                  END IF;

                  IF ( l_task.needs_analyze = 'Y' ) THEN
                     --Analyze
                     success := analyzepartitions( l_task ) AND success;
                  END IF;

               ELSE
                  -- UNKNOWN operation
                  success := FALSE;
            END CASE;

            IF ( success ) THEN

               -- Update the dangling task_status
               IF ( l_task.operation = g_MoveAnalyze_Op ) THEN
                  generic_end_task( l_task.job_id, l_task.task_id,
                    g_module_name||', Move and Analyze: '
                    || l_task.table_name||' done');
               END IF;

               update_part_task( l_task.job_id, l_task.task_id, g_OPS_SUCCESS );

               Log_Pkg.log_info_table(g_module_name, vprc, p_jobid, p_taskid,
                        SYSDATE, 'Move and Analyze: '|| l_task.table_name||' done');

            ELSE

               -- Update the dangling task_status
               IF ( l_task.operation = g_MoveAnalyze_Op ) THEN
                  generic_end_task( l_task.job_id, l_task.task_id,
                    g_module_name||', Move and Analyze: '
                    || l_task.table_name||' failed', FALSE);
               END IF;

               update_part_task( l_task.job_id, l_task.task_id, g_OPS_FAILED );

               Log_Pkg.log_warn_table(g_module_name, vprc, p_jobid, p_taskid,
                        SYSDATE, 'Operation '||l_task.operation||' on '|| l_task.table_name||' failed', NULL);

            END IF;

            COMMIT;

         END IF;

      END LOOP;

      -- finalize and quit
      Log_Pkg.log_info_table(g_module_name, vprc, p_jobid, p_taskid,
                        SYSDATE, 'Session '|| TO_CHAR(p_orajobid)||' completed');

      -- Finally
      COMMIT;

   END part_session;

   -- -----------------------------------------------------------------------
   -- testing
   -- -----------------------------------------------------------------------
   FUNCTION all_sessions_complete_x RETURN INTEGER
   IS

      l_return INTEGER := 0;

   BEGIN
      IF ( all_sessions_complete ) THEN
         l_return := 0;  -- TRUE
      ELSE
         l_return := 1;  -- FALSE
      END IF;

      RETURN l_return;

   END all_sessions_complete_x;

   -- -----------------------------------------------------------------------
   -- Get the next READY task
   -- -----------------------------------------------------------------------
   PROCEDURE get_next_task( p_jobid IN INTEGER,
                            p_task IN OUT NOCOPY taskRec )
   IS

      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT *
         FROM PART_TASK
         WHERE job_id = c_job_id
         AND status_code = c_code
         ORDER BY job_id, task_id, operation
         FOR UPDATE;

      task_rec task_cur%ROWTYPE;

      needs_task  BOOLEAN := TRUE;

      retry_count INTEGER := 0;
      max_retries INTEGER := 5;   -- This is not configurable, only the g_retry_interval

   BEGIN
      -- block until we have exclusive access...
      Lock_Util.get_x(g_task_lock_name);

      -- NULL the record
      p_task := NULL;

      -- There are 3 possible outcomes
      --   We acquire a task
      --   There are no more tasks available
      --   We fail to acquire a task due to a rule so we retry ... for how long?

      WHILE ( needs_task ) LOOP

         IF ( retry_count > 0 ) THEN
            dbms_lock.sleep( g_retry_interval );
         END IF;

         -- Try a single fetch ... hopefully we get a task
         OPEN task_cur( p_jobid, g_OPS_READY );
         FETCH task_cur INTO task_rec;

         IF ( task_cur%FOUND ) THEN

            -- *******************************************************************
            UPDATE PART_TASK
               SET status_code = g_OPS_ACTIVE
               WHERE CURRENT OF task_cur;
            -- *******************************************************************

            -- Build p_taskRec for return
            p_task.job_id           := task_rec.job_id;
            p_task.task_id          := task_rec.task_id;
            p_task.status_code      := task_rec.status_code;
            p_task.operation        := task_rec.operation;
            p_task.task_date        := task_rec.task_date;
            p_task.part_table_id    := task_rec.part_table_id;
            p_task.table_name       := task_rec.table_name;
            p_task.tablespace_name  := task_rec.tablespace_name;
            p_task.period           := task_rec.period;
            p_task.data_retention   := task_rec.data_retention;
            p_task.past_retention   := task_rec.past_retention;
            p_task.future_retention := task_rec.future_retention;
            p_task.needs_move       := task_rec.needs_move;
            p_task.needs_export     := task_rec.needs_export;
            p_task.needs_purge      := task_rec.needs_purge;
            p_task.needs_analyze    := task_rec.needs_analyze;
            p_task.analyze_mask     := task_rec.analyze_mask;
            p_task.part_name        := task_rec.part_name;

            CLOSE task_cur;
            needs_task := FALSE;

            COMMIT;

         -- We didn't find a task ... but why?
         ELSE

            CLOSE task_cur;

            -- Are there any tasks in the READY state?
            IF ( waiting_tasks( p_jobid ) ) THEN

               -- We must have hit a rule limitation so we retry ... later
               retry_count := retry_count + 1;

               -- Is is time to stop the looping?
               IF ( retry_count >= max_retries ) THEN
                  needs_task := FALSE;
               END IF;

            ELSE
               -- No tasks are available
               needs_task := FALSE;
            END IF;
         END IF;

      END LOOP;

      Lock_Util.release(g_task_lock_name);

      EXCEPTION
         WHEN OTHERS THEN
            IF ( task_cur%ISOPEN ) THEN
               CLOSE task_cur;
            END IF;

            Lock_Util.release(g_task_lock_name);

            RAISE;

   END get_next_task;

   -- -----------------------------------------------------------------------
   -- test
   -- -----------------------------------------------------------------------
   PROCEDURE get_next_task_x( p_jobid IN INTEGER )
   IS
      x taskRec;
   BEGIN
      get_next_task(p_jobid, x);

      IF ( x.task_id IS NULL ) THEN
         DBMS_OUTPUT.PUT_LINE('DONE');
      ELSE
         DBMS_OUTPUT.PUT_LINE(x.task_id);
      END IF;
   END get_next_task_x;

   -- -----------------------------------------------------------------------
   -- Update the part_task
   -- -----------------------------------------------------------------------
   PROCEDURE update_part_task( p_jobid IN INTEGER, p_taskid IN INTEGER, p_code IN INTEGER)
   IS
   BEGIN

      -- block until we have exclusive access...
      Lock_Util.get_x(g_task_lock_name);

      UPDATE PART_TASK
         SET status_code = p_code
      WHERE job_id = p_jobid
      AND task_id = p_taskid;

      Lock_Util.release(g_task_lock_name);

   EXCEPTION
      WHEN OTHERS THEN
         Lock_Util.release(g_task_lock_name);
         RAISE;

   END update_part_task;

   -- -----------------------------------------------------------------------
   -- Show how many tasks are remaining
   -- -----------------------------------------------------------------------
   PROCEDURE update_ready_count( p_jobid IN INTEGER )
   IS
      ready_count INTEGER := 0;
   BEGIN

      SELECT COUNT(*) INTO ready_count
      FROM PART_TASK
      WHERE job_id = p_jobid
      AND status_code = g_OPS_READY;

      dbms_application_info.set_client_info( '.... Tasks remaining: ' || TO_CHAR( ready_count) );

   END update_ready_count;

   -- -----------------------------------------------------------------------
   -- Poll while the sessions are running
   -- -----------------------------------------------------------------------
   FUNCTION task_poll( jobidparam IN INTEGER) RETURN INTEGER
   IS

      poll_jobs BOOLEAN := TRUE;

      -- Integer return code
      vreturn INTEGER := pm_success_c;

   BEGIN

      WHILE ( poll_jobs ) LOOP

         dbms_lock.sleep( g_poll_interval );

         update_ready_count( jobidparam );

         -- Have we been interrupted or have the part_tasks completed?
         IF ( job_interrupted( jobidparam ) ) THEN
            -- Mark all READY TASKS as ABORT so that we starve the sessions
            abort_tasks( jobidparam );

            vReturn := pm_failure_c;

            -- stop the loop
            poll_jobs := FALSE;

            -- We must commit so that the tasks reflect the ABORT
            -- and the Job status is is updated
            COMMIT;

         ELSE
            -- Are we complete?
            IF ( all_sessions_complete ) THEN
               -- stop the loop
               poll_jobs := FALSE;

               -- Did all of the tasks complete with a SUCCESS?
               IF ( all_tasks_success( jobidparam ) ) THEN
                  check_unmanaged_part_tables(jobidparam);
                  vReturn := pm_success_c;
               ELSE
                  vReturn := pm_failure_c;
               END IF;

            END IF;

         END IF;

      END LOOP;

      RETURN vReturn;

   END task_poll;

   -- -----------------------------------------------------------------------
   -- generic_start_task - A helper funciton to create a task for various
   -- jobs...anyone who wants to create a task on the fly.
   -- -----------------------------------------------------------------------
   FUNCTION generic_start_task(job_id INTEGER, msg VARCHAR2)
			       RETURN INTEGER  IS
      task_id INTEGER := NULL;
   BEGIN

      IF job_id IS NOT NULL THEN
	     INSERT INTO ACTIVE_TASK (task_id, job_id, sched_task_fk)
	     VALUES (seq_pm_task.NEXTVAL, job_id, -1)
	     RETURNING task_id INTO task_id;
      END IF;

      RETURN task_id;

   END generic_start_task;

   -- -----------------------------------------------------------------------
   -- generic_end_task - generic_start_task's mate
   -- -----------------------------------------------------------------------
   PROCEDURE generic_end_task(job_id INTEGER,
			      task_id INTEGER,
			      msg VARCHAR2,
			      success BOOLEAN DEFAULT TRUE) IS
      status_code INTEGER;
   BEGIN

      IF (success) THEN
	      status_code := g_OPS_SUCCESS;
      ELSE
	      status_code := g_OPS_FAILED;
      END IF;

   END generic_end_task;


   -- -----------------------------------------------------------------------
   -- Partition_Exists - Does the table/partition in question exist in the
   -- temp table?
   -- Querying via user_objects is *MUCH* faster than going to the more
   -- logical choice of USER_TAB_PARTITIONS.
   -- -----------------------------------------------------------------------
   FUNCTION partition_exists(table_name_in VARCHAR2,
        partition_name_in VARCHAR2) RETURN BOOLEAN IS
        ret_val BOOLEAN;
        ret_count INTEGER;
   BEGIN

      SELECT COUNT(*) INTO ret_count
        FROM user_objects
        WHERE     object_type = 'TABLE PARTITION'
        AND object_name = table_name_in
        AND  subobject_name = partition_name_in;

      IF (ret_count > 0) THEN
        ret_val := TRUE;
      ELSE
        ret_val := FALSE;
      END IF;

      RETURN ret_val;

   END partition_exists;

   -- -----------------------------------------------------------------------
   -- partition_date - Turn a partition into a date.  We should be able to
   -- do this natively in SQL but ORCL squawks.
   -- -----------------------------------------------------------------------
   FUNCTION partition_date(partition_name VARCHAR2) RETURN DATE IS
   BEGIN
      RETURN TO_DATE(Partition_Date_Str(partition_name), 'YYYYMMDDHH24MI');
   END partition_date;

   FUNCTION partition_date(partition_name VARCHAR2, p_mode CHAR) RETURN DATE IS
   BEGIN
     IF p_mode = 'L' THEN
       RETURN TO_DATE(Partition_Date_Str(partition_name), 'YYYYMMDDHH24MI');
     ELSE
       RETURN TO_DATE(Partition_Date_Str(partition_name), 'YYYYMMDDHH24');
     END IF;
   END partition_date;

   -- -----------------------------------------------------------------------
   -- partition_name - Turn a date into a partition name.
   -- -----------------------------------------------------------------------
   FUNCTION partition_name(partition_date DATE) RETURN VARCHAR2 IS
   BEGIN
      RETURN 'PART_'||TO_CHAR(partition_date, 'YYYYMMDD');
   END partition_name;

   FUNCTION partition_name(partition_date DATE, p_mode CHAR) RETURN VARCHAR2 IS
   BEGIN
     IF p_mode = 'L' THEN
       RETURN 'PART_'||TO_CHAR(partition_date, 'YYYYMMDDHH24MI');
     ELSE
       RETURN 'PART_'||TO_CHAR(partition_date, 'YYYYMMDD');
     END IF;
   END partition_name;

   -- -----------------------------------------------------------------------
   -- GetExtentSize - To a cached lookup of the tablespace's minextent length
   -- -----------------------------------------------------------------------
   FUNCTION getextentsize(tablespace_name_in VARCHAR2) RETURN INTEGER IS
      retval INTEGER;
   BEGIN

      IF (NVL(g_tablespace_name, tablespace_name_in || 'X') != tablespace_name_in)
      THEN
         SELECT tablespace_name, min_extlen INTO g_tablespace_name, g_extent_size
           FROM user_tablespaces
           WHERE tablespace_name = tablespace_name_in;
      END IF;

      retval := g_extent_size;

      RETURN retval;

   END getextentsize;

   PROCEDURE check_unmanaged_part_tables(jobidparam IN INTEGER) IS

      CURSOR c_unmanaged_tabs IS
       SELECT DISTINCT u.table_name table_name
        FROM user_tab_partitions u, PART_TABLES p
        WHERE p.table_name(+)=u.table_name
        AND high_value IS NOT NULL
        AND p.table_name IS NULL;

      task_id NUMBER;
   BEGIN
      FOR c_unmanaged_tabs_rec IN c_unmanaged_tabs LOOP
         Log_Pkg.log_info_table(g_module_name, NULL, NULL, NULL, SYSDATE,
                 c_unmanaged_tabs_rec.table_name||' is not manage by Partition Maintenance');
      END LOOP;

   END check_unmanaged_part_tables;

   -- -----------------------------------------------------------------------
   -- CheckSpace - Check to ensure there is sufficient space in the target
   -- tablespace for the required number of partitions for a particular table
   -- -----------------------------------------------------------------------
   FUNCTION CheckSpace(table_name_in VARCHAR2, tablespace_name_in VARCHAR2, data_retention_in INTEGER,
               future_retention_in INTEGER, checkparam VARCHAR2) RETURN INTEGER IS
      retval            INTEGER := 0;
      tbs_free_space    INTEGER := 0;
      tbs_req_space     INTEGER := 0;
      num_req_part      INTEGER := 0;

      vreturn INTEGER := pm_success_c;
      vprc VARCHAR2(30) := 'CheckSpace';

      CURSOR c_tablespace_list IS
          SELECT pt.tablespace_name, SUM(fs.bytes) free_space
          FROM dba_free_space fs, PART_TABLES pt
          WHERE fs.tablespace_name = pt.tablespace_name
          GROUP BY pt.tablespace_name;

      tbs_rec c_tablespace_list%ROWTYPE;

   BEGIN

      OPEN c_tablespace_list;
      FETCH c_tablespace_list INTO tbs_rec;

      IF ( c_tablespace_list%FOUND ) THEN

         g_TbsSpace(tbs_rec.tablespace_name)   := tbs_rec.free_space;

         CLOSE c_tablespace_list;
      END IF;

      -- Select the free space in the target tablespace
      SELECT SUM(bytes) total_bytes
      INTO tbs_free_space
      FROM dba_free_space
      WHERE tablespace_name=tablespace_name_in;


      -- Select the required number of partitions for the target table
      IF ( checkparam = 'PushAhead' ) THEN

          SELECT TRUNC((TRUNC(SYSDATE) + future_retention_in) - (MAX(TO_DATE (Partition_Date_Str(partition_name), 'YYYYMMDDHH24'))))
          INTO num_req_part
          FROM   user_tab_partitions
          WHERE  table_name = table_name_in;

      ELSIF (checkparam = 'PushBehind' ) THEN

          SELECT TRUNC((TRUNC(SYSDATE) + data_retention_in) - (MAX(TO_DATE (Partition_Date_Str(partition_name), 'YYYYMMDDHH24MI'))))
          INTO num_req_part
          FROM   user_tab_partitions
          WHERE  table_name = table_name_in;

      END IF;

      -- select the required space for the tables partitions
      SELECT ( NVL(next_extent,0) * NVL(num_req_part,0) )
      INTO tbs_req_space
      FROM   user_tablespaces
      WHERE tablespace_name = tablespace_name_in;

      IF ( NVL(tbs_free_space,0)  > NVL(tbs_req_space,0) ) THEN
         retval := 0;
      ELSE

         Log_Pkg.log_warn_table(g_module_name, vprc, NULL, NULL, SYSDATE,
                 'Insufficient space for '||table_name_in||' in '
                 ||tablespace_name_in||', required : '||ROUND(tbs_req_space/1024/1024)||'MB', NULL);

         retval := ROUND(tbs_req_space/1024/1024);
      END IF;

      RETURN retval;

   END checkspace;


   PROCEDURE get_pm_sessions(results OUT t_ref_cursor)
   IS

    BEGIN

        OPEN results FOR
            SELECT VALUE FROM ADM_SYSTEM_VARIABLES WHERE NAME='PartMaintSessions';

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_pm_sessions;

   PROCEDURE update_pm_sessions(l_value IN VARCHAR2)
   IS

       retval      BOOLEAN        := TRUE;
       retmsg      VARCHAR2(9000) := NULL;
       l_err_msg   VARCHAR2 (500);
       l_str       VARCHAR2(1024);

   BEGIN

       l_str := 'UPDATE adm_system_variables SET value '
               ||' = '||l_value
               ||' WHERE name = ''PartMaintSessions''';

       BEGIN
          DBMS_OUTPUT.PUT_LINE(l_str);
          EXECUTE IMMEDIATE l_str;
       EXCEPTION WHEN OTHERS THEN
            IF retval THEN
              retmsg := SQLERRM;
          END IF;

          Log_Pkg.warn(g_module_name, 'error updating session value. '||SQLERRM);
          retval := FALSE;
       END;

   EXCEPTION WHEN OTHERS THEN
      RAISE;

   END update_pm_sessions;


   -- -----------------------------------------------------------------------
   -- TableType Exists - Check to see if table type exist in the subclass table
   -- -----------------------------------------------------------------------
   FUNCTION tabletypeexists(type_name_in VARCHAR2,
        subtype_name_in VARCHAR2 DEFAULT NULL) RETURN BOOLEAN IS
        ret_val BOOLEAN;
        ret_count INTEGER;
   BEGIN

      IF (subtype_name_in IS NULL)
      THEN

          SELECT COUNT(*) INTO ret_count
            FROM part_maint_types_vw
            WHERE UPPER(TYPE) = UPPER(type_name_in);

      ELSE

          SELECT COUNT(*) INTO ret_count
            FROM part_maint_types_vw
            WHERE UPPER(TYPE) = UPPER(type_name_in)
            AND UPPER(NAME) = UPPER(subtype_name_in);

      END IF;

      IF (ret_count > 0) THEN
        ret_val := TRUE;
      ELSE
        ret_val := FALSE;
      END IF;

      RETURN ret_val;

   END tabletypeexists;


   PROCEDURE get_space_usage(results OUT t_ref_cursor)
   IS

    BEGIN

        OPEN results FOR
            SELECT ts.tablespace_name      tablespace_name ,
                   ts.ts_size/ (1024*1024) total_mb ,
                   DECODE(SIGN((free.ts_free/ts.ts_size)*100 - 10), 1,' ',' <**> ') over_flag ,
                   ROUND((free.ts_free/ts.ts_size)*100,1)               pct_free ,
                   ROUND(free.ts_free/(1024*1024),1)                     free_mb  ,
                   ROUND((ts.ts_size - free.ts_free)/(1024*1024),1)        used_mb
            FROM
                  (SELECT tablespace_name, SUM(bytes) ts_size
                   FROM dba_data_files
                   GROUP BY tablespace_name
                  ) ts,
                  (SELECT df.tablespace_name, SUM(NVL(fs.bytes, 0)) ts_free
                   FROM sys.dba_free_space fs, sys.dba_data_files df
                   WHERE fs.file_id (+)= df.file_id
                   GROUP BY df.tablespace_name
                  ) free
            WHERE ts.tablespace_name = free.tablespace_name
            ORDER BY ts.tablespace_name;

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_space_usage;


   PROCEDURE get_valid_types(l_type IN VARCHAR2, l_subtype IN VARCHAR2, results OUT t_ref_cursor)
   IS

    BEGIN

        OPEN results FOR
            SELECT TYPE, NAME
            FROM part_maint_types_vw
            WHERE
               UPPER(TYPE) LIKE NVL(UPPER(DECODE (l_type,'ALL',NULL,l_type)),'%')
            AND
               UPPER(NAME) LIKE NVL(UPPER(DECODE (l_subtype,'ALL',NULL,l_subtype)),'%');

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_valid_types;

   PROCEDURE get_main_types(l_type IN VARCHAR2, results OUT t_ref_cursor)
   IS

    BEGIN

       IF ( tabletypeexists( l_type))
       THEN

           OPEN results FOR
               SELECT NAME
               FROM part_maint_types_vw
               WHERE
                   UPPER(TYPE) LIKE UPPER(l_type)
               ORDER BY NAME ASC;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_main_types;

   PROCEDURE get_list_partitions(l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                             l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                             l_sdate_in IN VARCHAR2 DEFAULT NULL, l_edate_in IN VARCHAR2 DEFAULT NULL,
                             results OUT t_ref_cursor)
   IS
      l_sdate  DATE;
      l_edate  DATE;
      invalid_date_err EXCEPTION;

    BEGIN

       --Check for the validity of the dates
       IF (l_sdate_in IS NOT NULL)
       THEN
           BEGIN
               SELECT TO_DATE(l_sdate_in,'yyyymmddHH24')
               INTO l_sdate
               FROM dual;
           EXCEPTION WHEN OTHERS THEN
               RAISE invalid_date_err;
           END;
       END IF;

       IF (l_edate_in IS NOT NULL)
       THEN
           BEGIN
               SELECT TO_DATE(l_edate_in,'yyyymmddHH24')
               INTO l_edate
               FROM dual;
           EXCEPTION WHEN OTHERS THEN
               RAISE invalid_date_err;
           END;
       END IF;


       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           OPEN results FOR
               SELECT DISTINCT v.table_name, v.partition_name, DECODE(p.part_eff_date,NULL,'-','PINNED') "STATUS"
               FROM PART_MAINT_PART_DETAILS_VW v, PART_PINNED p
               WHERE
                   v.part_table_id = p.part_table_id (+)
               AND
                   v.partition_date = p.part_eff_date (+)
               AND
                   UPPER(v.TYPE)       LIKE NVL(UPPER(l_type),'%')
               AND
                   UPPER(v.sub_type)   LIKE NVL(UPPER(l_sub_type),'%')
               AND
                   UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type),'%')
               AND
                   UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name),'%')
               AND
                   v.partition_name  >= Part_Maint.partition_name(NVL(l_sdate,TO_DATE('197001010000','YYYYMMDDHH24MI')))
               AND
                   v.partition_name  <= Part_Maint.partition_name(NVL(l_edate, TO_DATE('203801010000','YYYYMMDDHH24MI')))
               ORDER BY v.table_name, v.partition_name;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_list_partitions;


   PROCEDURE get_summary_partitions(l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                             l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                             results OUT t_ref_cursor)
   IS

    BEGIN

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           OPEN results FOR
               SELECT v.table_name, v.table_type, v.min_part_name, v.max_part_name, v.count_partitions
               FROM PART_MAINT_PART_SUMMARY_VW v
               WHERE
                   UPPER(v.TYPE) LIKE NVL(UPPER(l_type),'%')
               AND
                   UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type),'%')
               AND
                   UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type),'%')
               AND
                   UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name),'%')
               ORDER BY v.table_name, v.max_part_name;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

    EXCEPTION WHEN OTHERS THEN
       RAISE;

    END get_summary_partitions;

  PROCEDURE get_pinned_partitions(l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                             l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                             l_sdate_in IN VARCHAR2 DEFAULT NULL, l_edate_in IN VARCHAR2 DEFAULT NULL,
                             results OUT t_ref_cursor)
   IS
      l_sdate  DATE;
      l_edate  DATE;
      invalid_date_err EXCEPTION;

    BEGIN

       --Check for the validity of the dates
       IF (l_sdate_in IS NOT NULL)
       THEN
           BEGIN
               SELECT TO_DATE(l_sdate_in,'yyyymmddHH24')
               INTO l_sdate
               FROM dual;
           EXCEPTION WHEN OTHERS THEN
               RAISE invalid_date_err;
           END;
       END IF;

       IF (l_edate_in IS NOT NULL)
       THEN
           BEGIN
               SELECT TO_DATE(l_edate_in,'yyyymmddHH24')
               INTO l_edate
               FROM dual;
           EXCEPTION WHEN OTHERS THEN
               RAISE invalid_date_err;
           END;
       END IF;

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           OPEN results FOR
               SELECT table_name, part_eff_date
               FROM part_maint_pinned_details_vw
               WHERE
                   UPPER(TYPE) LIKE NVL(UPPER(l_type),'%')
               AND UPPER(sub_type) LIKE NVL(UPPER(l_sub_type),'%')
               AND UPPER(table_type) LIKE NVL(UPPER(l_tab_type),'%')
               AND UPPER(table_name) LIKE NVL(UPPER(l_filter_name),'%')
               AND partition_name >= Part_Maint.partition_name(NVL(l_sdate,TO_DATE('197001010000','YYYYMMDDHH24MI')))
               AND partition_name <= Part_Maint.partition_name(NVL(l_edate,TO_DATE('203801010000','YYYYMMDDHH24MI')))
               ORDER BY table_name, part_eff_date ;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pinned_partitions;

   PROCEDURE get_parameters (l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                             l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                             results OUT t_ref_cursor)
   IS

   BEGIN

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           OPEN results FOR
               SELECT DISTINCT
                   table_name, table_type, tablespace_name,
                   period, NVL(data_retention,'0'), NVL(past_retention,'0'),
                   NVL(future_retention,'0'), needs_move, needs_export, needs_purge,
                   needs_analyze, analyze_mask
               FROM part_maint_param_details_vw
               WHERE
                   UPPER(TYPE) LIKE NVL(UPPER(l_type),'%')
               AND UPPER(sub_type) LIKE NVL(UPPER(l_sub_type),'%')
               AND UPPER(table_type) LIKE NVL(UPPER(l_tab_type),'%')
               AND UPPER(table_name) LIKE NVL(UPPER(l_filter_name),'%')
               ORDER BY table_type, table_name;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_parameters;


   PROCEDURE get_default_parameters (results OUT t_ref_cursor)
   IS
   BEGIN

       OPEN results FOR
           SELECT DISTINCT
              table_type, period, NVL(data_retention,'0'), NVL(past_retention,'0'),
              NVL(future_retention,'0'), COUNT(*) numtables
           FROM part_maint_param_details_vw
           WHERE TYPE='TABLE'
           GROUP BY table_type, period, NVL(data_retention,'0'),
                NVL(past_retention,'0'), NVL(future_retention,'0')
           ORDER BY table_type;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_default_parameters;

   PROCEDURE get_analyze_mask_details (results OUT t_ref_cursor)
   IS
        l_day NUMBER;
   BEGIN

        OPEN results FOR
            SELECT 'StartOfWeek', TO_CHAR(TRUNC(SYSDATE,'D'),'DAY','NLS_DATE_LANGUAGE=AMERICAN') FROM dual;

   END get_analyze_mask_details;

   PROCEDURE update_parameters (l_param IN VARCHAR2, l_value IN VARCHAR2, l_type IN VARCHAR2,
                                l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                                l_filter_name IN VARCHAR2 DEFAULT NULL)
   IS

       retval      BOOLEAN        := TRUE;
       retmsg      VARCHAR2(9000) := NULL;
       l_err_msg   VARCHAR2 (500);
       l_str       VARCHAR2(1024);

   BEGIN

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN
           l_str := 'UPDATE part_tables SET '||l_param
                   ||' = UPPER('''||l_value||''')'
                   ||' WHERE table_name IN'
                   ||' (SELECT DISTINCT table_name '
                   ||' FROM part_maint_param_details_vw '
                   ||' WHERE UPPER(type) like NVL(UPPER('''||l_type||'''),''%'')'
                   ||' AND UPPER(sub_type) like NVL(UPPER('''||l_sub_type||'''),''%'')'
                   ||' AND UPPER(table_type) like NVL(UPPER('''||l_tab_type||'''),''%'')'
                   ||' AND UPPER(table_name) like NVL(UPPER('''||l_filter_name||'''),''%''))';

           BEGIN
              DBMS_OUTPUT.PUT_LINE(l_str);
              EXECUTE IMMEDIATE l_str;
              COMMIT;
           EXCEPTION WHEN OTHERS THEN
              IF retval THEN
                  retmsg := SQLERRM;
              END IF;

              l_err_msg := 'Error Updating Parameters,'||l_param||' for '||l_type||','||l_sub_type;
              DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.UPDATE_PARAMETERS:'||l_err_msg);
              DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||SQLERRM);
              retval := FALSE;
              RAISE;
           END;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   END update_parameters;

   -------------------------------------------------------------------------
   -- get_pm_logs - List logs associated with the pm process
   -------------------------------------------------------------------------
  PROCEDURE get_pm_logs(results OUT t_ref_cursor)
   IS

   BEGIN

           OPEN results FOR
               SELECT DISTINCT job_id,
                   NVL(TO_CHAR(MIN(log_time),'YYYY/MM/DD HH24:MI'),'-'),
                   NVL(TO_CHAR(MAX(log_time),'YYYY/MM/DD HH24:MI'),'-'),
                   NVL(ROUND((MAX(log_time) - MIN(log_time))*24*60*60,2),'0')
               FROM ADM_LOG
               WHERE event_name = 'PartMaint'
               HAVING MIN(log_time) > SYSDATE-30
               GROUP BY job_id
               ORDER BY NVL(TO_CHAR(MIN(log_time),'YYYY/MM/DD HH24:MI'),'-');

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pm_logs;

   -------------------------------------------------------------------------
   -- get_pm_latest_logs - List detailed logs for latest process
   -------------------------------------------------------------------------
  PROCEDURE get_pm_latest_logs(results OUT t_ref_cursor)
   IS

   BEGIN

           OPEN results FOR
               SELECT
                   NVL(object_name,'PartMaint'),
                   NVL(TO_CHAR(log_time,'YYYY/MM/DD HH24:MI'),'-'),
                   NVL(SUBSTR(message,0,150),'-')
               FROM ADM_LOG
               WHERE event_name='PartMaint'
               AND job_id =
                    (SELECT DISTINCT job_id
                     FROM ADM_LOG
                     WHERE log_time = (SELECT MAX(log_time) FROM ADM_LOG WHERE event_name='PartMaint')
                     AND event_name='PartMaint')
               ORDER BY log_time;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pm_latest_logs;

   -------------------------------------------------------------------------
   -- get_pm_detailed_logs - List detailed logs associated with a specified pm process
   -------------------------------------------------------------------------
  PROCEDURE get_pm_detailed_logs(l_log_id IN VARCHAR2, results OUT t_ref_cursor)
   IS

    BEGIN

           OPEN results FOR
               SELECT
                   NVL(object_name,'PartMaint'),
                   NVL(TO_CHAR(log_time,'YYYY/MM/DD HH24:MI'),'-'),
                   NVL(SUBSTR(message,0,180),'-')
               FROM ADM_LOG
               WHERE event_name='PartMaint'
               AND job_id = TO_NUMBER(l_log_id)
               ORDER BY log_time;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pm_detailed_logs;

   -------------------------------------------------------------------------
   -- get_pm_error_logs - List error logs associated with a specified pm process
   -------------------------------------------------------------------------
  PROCEDURE get_pm_error_logs(results OUT t_ref_cursor)
   IS

    BEGIN
           OPEN results FOR
               SELECT
               NVL(job_id,'0'),
               NVL(TO_CHAR(log_time,'YYYY/MM/DD HH24:MI'),'-'),
               NVL(SUBSTR(message,0,150),'-')
               FROM ADM_LOG
               WHERE ( UPPER(message) LIKE UPPER('%ORA-%')
               OR UPPER(message) LIKE UPPER('%failed%')
               OR UPPER(message) LIKE UPPER('%exception%')
               OR UPPER(message) LIKE UPPER('%error%'))
               AND log_time > SYSDATE-60
               ORDER BY log_time;

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pm_error_logs;

   -------------------------------------------------------------------------
   -- get_pm_status - List process status
   -------------------------------------------------------------------------
  PROCEDURE get_pm_status(results OUT t_ref_cursor)
   IS

    BEGIN

           OPEN results FOR
               SELECT client_info, module, NVL(action,'-'), SID, process, status
               FROM v$session
               WHERE module='PartMaint';

   EXCEPTION WHEN OTHERS THEN
       RAISE;

   END get_pm_status;

   -------------------------------------------------------------------------
   -- alter_partition_short - Adds partitions for the given table ... ...
   -------------------------------------------------------------------------
   PROCEDURE alter_partition_short(p_table_name VARCHAR2,
                            p_new_part VARCHAR2,
                            p_new_part_end_period DATE,
                            p_tablespace_name VARCHAR2) IS
   BEGIN

        EXECUTE IMMEDIATE ('ALTER TABLE ' || p_table_name
                                      || ' ADD PARTITION ' || p_new_part
                                      || ' VALUES LESS THAN (TO_DATE(' || CHR (39)
                                      || TO_CHAR (p_new_part_end_period, 'YYYYMMDD HH24:MI:SS')
                                      || CHR (39) || ',' || CHR (39)
                                      || 'YYYYMMDD HH24:MI:SS'
                                      || CHR (39) || ')) TABLESPACE '
                                      || p_tablespace_name );

       EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_new_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';

   EXCEPTION
    WHEN OTHERS THEN
    RAISE;
    END   alter_partition_short;

   -------------------------------------------------------------------------
   -- alter_partition_long - Adds partitions for the given table ... ...
   -------------------------------------------------------------------------
   PROCEDURE alter_partition_long(p_table_name VARCHAR2,
                            p_new_part VARCHAR2,
                            p_new_part_end_period DATE,
                            p_tablespace_name VARCHAR2) IS
   BEGIN

        EXECUTE IMMEDIATE ('ALTER TABLE ' || p_table_name
                                      || ' ADD PARTITION ' || p_new_part
                                      || ' VALUES LESS THAN ('||TO_NUMBER(TO_CHAR(p_new_part_end_period, 'YYYYMMDDHH24MI'))||') TABLESPACE '
                                      || p_tablespace_name );

       EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_new_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';

   EXCEPTION
    WHEN OTHERS THEN
    RAISE;
    END   alter_partition_long;
    
 
   -------------------------------------------------------------------------
   -- split_partition_short - Adds partitions for the given table ... ...
   -------------------------------------------------------------------------
   PROCEDURE split_partition_short(p_table_name VARCHAR2, p_previous_part VARCHAR2,
                            p_new_part VARCHAR2, p_new_part_end_period DATE,
                            p_previous_tablespace_name VARCHAR2, p_tablespace_name VARCHAR2) IS
    attempt_count    INTEGER        := 0;
    resource_busy EXCEPTION;
    sql_statement VARCHAR2(1000);
    PRAGMA EXCEPTION_INIT( resource_busy, -00054 );
   BEGIN

        WHILE (attempt_count < g_resource_attempts )
        LOOP

            BEGIN
                -- Lock the partition for the curdate_name
                lock_table(p_table_name,p_previous_part);


                sql_statement := 'ALTER TABLE '|| p_table_name
                                      || ' split PARTITION '||p_previous_part
                                      || ' AT ( TO_DATE('
                                      || '''' || TO_CHAR(p_new_part_end_period, 'YYYY/MM/DD HH24') || ''''
                                      || ', ' || '''' || 'YYYY/MM/DD HH24' || ''''
                                      || ')) INTO (PARTITION ' || p_new_part
                                      || ' TABLESPACE ' || p_tablespace_name
                                      || ', PARTITION ' || p_previous_part;

                IF p_previous_tablespace_name IS NOT NULL THEN
                       sql_statement := sql_statement || ' TABLESPACE ' || p_previous_tablespace_name;
                END IF;

                sql_statement := sql_statement || ')';

                EXECUTE IMMEDIATE (sql_statement);

               EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_previous_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
               EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_new_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';

               EXIT;

            EXCEPTION
                WHEN resource_busy THEN
                    attempt_count := attempt_count + 1;
                    dbms_application_info.set_client_info( p_table_name ||
                           ': retry ' || TO_CHAR(attempt_count) );
                           dbms_lock.sleep(g_resource_interval);
                WHEN OTHERS THEN
                    RAISE;
            END;
        END LOOP;

   EXCEPTION
    WHEN OTHERS THEN
        RAISE;
   END split_partition_short;
                           
   -------------------------------------------------------------------------
   -- split_partition_long - Adds partitions for the given table ... ...
   -------------------------------------------------------------------------
   PROCEDURE split_partition_long(p_table_name VARCHAR2, p_previous_part VARCHAR2,
                            p_new_part VARCHAR2, p_new_part_end_period DATE,
                            p_previous_tablespace_name VARCHAR2, p_tablespace_name VARCHAR2) IS
    attempt_count    INTEGER        := 0;
    resource_busy EXCEPTION;
    sql_statement VARCHAR2(1000);
    PRAGMA EXCEPTION_INIT( resource_busy, -00054 );
   BEGIN

        WHILE (attempt_count < g_resource_attempts )
        LOOP

            BEGIN
                -- Lock the partition for the curdate_name
                lock_table(p_table_name,p_previous_part);
                                                  
                sql_statement := 'ALTER TABLE '|| p_table_name
                                      || ' split PARTITION '||p_previous_part
                                      || ' AT ('|| TO_NUMBER(TO_CHAR(p_new_part_end_period, 'YYYYMMDDHH24MI'))
                                      || ') INTO (PARTITION ' || p_new_part
                                      || ' TABLESPACE ' || p_tablespace_name
                                      || ', PARTITION ' || p_previous_part;



                IF p_previous_tablespace_name IS NOT NULL THEN
                       sql_statement := sql_statement || ' TABLESPACE ' || p_previous_tablespace_name;
                END IF;

                sql_statement := sql_statement || ')';

                EXECUTE IMMEDIATE (sql_statement);

               EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_previous_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';
               EXECUTE IMMEDIATE 'ALTER TABLE '|| p_table_name ||' MODIFY PARTITION '
                                 || p_new_part
                                 ||' REBUILD UNUSABLE LOCAL INDEXES';

               EXIT;

            EXCEPTION
                WHEN resource_busy THEN
                    attempt_count := attempt_count + 1;
                    dbms_application_info.set_client_info( p_table_name ||
                           ': retry ' || TO_CHAR(attempt_count) );
                           dbms_lock.sleep(g_resource_interval);
                WHEN OTHERS THEN
                    RAISE;
            END;
        END LOOP;

   EXCEPTION
    WHEN OTHERS THEN
        RAISE;
   END split_partition_long;


   ------------------------------------------------------------------------------------------------------
   -- pin_paritions - Pin partitions in the database so they are not dropped by the scheduled part_maint
   ------------------------------------------------------------------------------------------------------`
   PROCEDURE pin_partitions(l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                            l_sub_type IN VARCHAR2 DEFAULT NULL,l_tab_type IN VARCHAR2 DEFAULT NULL,
                            l_filter_name IN VARCHAR2 DEFAULT NULL)
   IS

      invalid_date_err EXCEPTION;
      l_sdate          DATE;
      l_edate          DATE;
      retval           BOOLEAN        := TRUE;
      retmsg           VARCHAR2(9000) := NULL;
      vprc             VARCHAR2(30)   := 'PinPartition';
      l_str            VARCHAR2 (1000);
      l_err_msg        VARCHAR2(200)  := NULL;


   BEGIN

       --Check for the validity of the dates
       BEGIN
           SELECT TO_DATE(l_sdate_in,'YYYYMMDDHH24'),TO_DATE(l_edate_in,'YYYYMMDDHH24')
           INTO l_sdate,l_edate
           FROM dual;
       EXCEPTION WHEN OTHERS THEN
           RAISE invalid_date_err;
       END;


       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           l_str := 'INSERT INTO part_pinned '
                    ||' SELECT part_table_id, partition_date '
                    ||' FROM PART_MAINT_PART_DETAILS_VW '
                    ||' WHERE UPPER(type) like nvl(UPPER('''||l_type||'''),''%'')  '
                    ||' AND UPPER(sub_type) like nvl(UPPER('''||l_sub_type||'''),''%'')  '
                    ||' AND UPPER(table_type) like nvl(UPPER('''||l_tab_type||'''),''%'') '
                    ||' AND UPPER(table_name) like nvl(UPPER('''||l_filter_name||'''),''%'') '
                    ||' AND partition_name >= '''||Part_Maint.partition_name(l_sdate)||''''
                    ||' AND partition_name <= '''||Part_Maint.partition_name(l_edate)||''''
                    ||' MINUS '
                    ||' select part_table_id, part_eff_date  '
                    ||' FROM PART_MAINT_PINNED_DETAILS_VW '
                    ||' WHERE UPPER(type) like nvl(UPPER('''||l_type||'''),''%'') '
                    ||' AND UPPER(sub_type) like nvl(UPPER('''||l_sub_type||'''),''%'') '
                    ||' AND UPPER(table_type) like nvl(UPPER('''||l_tab_type||'''),''%'') '
                    ||' AND UPPER(table_name) like nvl(UPPER('''||l_filter_name||'''),''%'') '
                    ||' AND partition_name >= '''||Part_Maint.partition_name(l_sdate)||''''
                    ||' AND partition_name <= '''||Part_Maint.partition_name(l_edate)||'''';

           BEGIN
              DBMS_OUTPUT.PUT_LINE(l_str);
              EXECUTE IMMEDIATE l_str;
              COMMIT;
           EXCEPTION WHEN OTHERS THEN
              IF retval THEN
                  retmsg := SQLERRM;
              END IF;

              l_err_msg := 'Error Pinning Partitions, for '||l_type||','||l_sub_type||','||l_tab_type;
              DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.PIN_PARTITIONS:'||l_err_msg);
              DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||retmsg);
              retval := FALSE;
              RAISE;

           END;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   END pin_partitions;

   ------------------------------------------------------------------------------------------------------
   -- pin_paritions - Pin partitions in the database so they are not dropped by the scheduled part_maint
   ------------------------------------------------------------------------------------------------------`
   PROCEDURE unpin_partitions(l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                            l_sub_type IN VARCHAR2 DEFAULT NULL,l_tab_type IN VARCHAR2 DEFAULT NULL,
                            l_filter_name IN VARCHAR2 DEFAULT NULL)
   IS

         CURSOR cur_pinned_details(l_start_date DATE, l_end_date DATE,
                               l_type_in VARCHAR2, l_sub_type_in VARCHAR2 DEFAULT NULL,
                               l_tab_type_in VARCHAR2 DEFAULT NULL, l_filter_name_in VARCHAR2 DEFAULT NULL)  IS
            SELECT part_table_id, part_eff_date
            FROM PART_MAINT_PINNED_DETAILS_VW
            WHERE
                UPPER(TYPE) LIKE NVL(UPPER(l_type_in),'%')
            AND UPPER(sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
            AND UPPER(table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
            AND UPPER(table_name) LIKE NVL(UPPER(l_filter_name_in),'%')
            AND part_eff_date >= l_start_date
            AND part_eff_date <= l_end_date;

      invalid_date_err EXCEPTION;
      l_sdate          DATE;
      l_edate          DATE;
      retval           BOOLEAN        := TRUE;
      retmsg           VARCHAR2(9000) := NULL;
      vprc             VARCHAR2(30)   := 'UnpinPartition';
      l_str            VARCHAR2 (1000);
      l_err_msg        VARCHAR2(200)  := NULL;


   BEGIN

       DBMS_OUTPUT.PUT_LINE('Starting unpinning of partitions');

       --Check for the validity of the dates
       BEGIN
           SELECT TO_DATE(l_sdate_in,'YYYYMMDDHH24'),TO_DATE(l_edate_in,'YYYYMMDDHH24')
           INTO l_sdate,l_edate
           FROM dual;
       EXCEPTION WHEN OTHERS THEN
           RAISE invalid_date_err;
       END;

       DBMS_OUTPUT.PUT_LINE('....dates have been set');
       IF ( tabletypeexists( l_type, l_sub_type))
       THEN
            FOR l_pin_rec IN cur_pinned_details(l_sdate, l_edate, l_type, l_sub_type, l_tab_type, l_filter_name) LOOP

                  l_str := 'DELETE FROM part_pinned where part_table_id='
                           || l_pin_rec.part_table_id
                           || ' and part_eff_date = '''
                           || l_pin_rec.part_eff_date||'''';

                   BEGIN
                      EXECUTE IMMEDIATE l_str;
                      COMMIT;
                   EXCEPTION WHEN OTHERS THEN
                      IF retval THEN
                          retmsg := SQLERRM;
                      END IF;

                      l_err_msg := 'Error Unpinning Partitions, for '||l_type||','||l_sub_type||','||l_tab_type;
                      DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.UNPIN_PARTITIONS:'||l_err_msg);
                      DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||retmsg);
                      retval := FALSE;
                      RAISE;
                   END;

            END LOOP;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   END unpin_partitions;

   -------------------------------------------------------------------------
   -- add_partitions - Adds partitions for the given date range.
   -------------------------------------------------------------------------
   PROCEDURE add_partitions(l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                            l_sub_type IN VARCHAR2 DEFAULT NULL,l_tab_type IN VARCHAR2 DEFAULT NULL,
                            l_filter_name IN VARCHAR2 DEFAULT NULL)
   IS

      CURSOR cur_table_details(l_type_in VARCHAR2, l_sub_type_in VARCHAR2 DEFAULT NULL,
                     l_tab_type_in VARCHAR2 DEFAULT NULL, l_filter_name_in VARCHAR2 DEFAULT NULL)  IS
         SELECT TYPE, sub_type, table_type, part_table_id, table_name, tablespace_name,
             period, data_retention, past_retention, future_retention, min_value, max_value
         FROM PART_MAINT_PART_SUMMARY_VW
         WHERE UPPER(TYPE) LIKE NVL(UPPER(l_type_in),'%')
         AND UPPER(sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(table_name) LIKE NVL(UPPER(l_filter_name_in),'%');

      invalid_date_err EXCEPTION;

      overwrite_partition_err EXCEPTION;
      PRAGMA EXCEPTION_INIT(overwrite_partition_err, -14074);
      partition_bounds_err EXCEPTION;
      PRAGMA EXCEPTION_INIT(partition_bounds_err, -14080);

      retmsg           VARCHAR2(9000) := NULL;
      vprc             VARCHAR2(30)   := 'AddPartition';

      l_sdate              DATE;
      l_edate              DATE;

      l_current_max_part   DATE;
      l_target_max_part    DATE;
      l_current_min_part   DATE;
      l_target_min_part    DATE;
      l_retention_min_part DATE;

      l_local_sql           VARCHAR2(1000);
      l_err_msg           VARCHAR2 (500);

   BEGIN

       --Check for the validity of the dates
       BEGIN
           SELECT TO_DATE(l_sdate_in,'yyyymmddHH24'),TO_DATE(l_edate_in,'yyyymmddHH24')
           INTO l_sdate,l_edate
           FROM dual;
       EXCEPTION WHEN OTHERS THEN
           RAISE invalid_date_err;
       END;

      DBMS_OUTPUT.PUT_LINE('l_sdate_in:'||l_sdate_in);
      DBMS_OUTPUT.PUT_LINE('l_edate_in:'||l_edate_in);
      DBMS_OUTPUT.PUT_LINE('l_type:'||l_type);
      DBMS_OUTPUT.PUT_LINE('l_sub_type:'||l_sub_type);
      DBMS_OUTPUT.PUT_LINE('l_tab_type:'||l_tab_type);

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN

           FOR l_tab_rec IN cur_table_details(l_type, l_sub_type, l_tab_type, l_filter_name) LOOP

               l_current_max_part   := TRUNC (l_tab_rec.max_value);
               l_current_min_part   := TRUNC (l_tab_rec.min_value);
               l_target_max_part    := Adm_Date.trunc_to_last_period(l_edate, l_tab_rec.period);
               l_target_min_part    := l_sdate;
               l_retention_min_part := Adm_Date.add_periods(Adm_Date.trunc_to_last_period(SYSDATE, l_tab_rec.period),
                                                l_tab_rec.period,-(l_tab_rec.data_retention));

               DBMS_OUTPUT.PUT_LINE(l_tab_rec.table_name||', curr min:'||l_current_min_part||', curr max:'||l_current_max_part);
               DBMS_OUTPUT.PUT_LINE('... target min:'||l_target_min_part||', target max:'||l_target_max_part);

               -- loop through the partition we need to create
               WHILE l_target_max_part >= l_target_min_part LOOP

                   IF NOT l_tab_rec.period = 'Q' THEN
                    BEGIN
                        SELECT DECODE(NVL(previous_part,'not_found'),'not_found',
                                      'part_maint.alter_partition_short(''' || table_name || ''',''PART_'
                                            || TO_CHAR(l_target_max_part,'YYYYMMDD') || ''',to_date('''
                                            || TO_CHAR(new_part_end_period,'yyyymmddhh24') || ''',''yyyymmddhh24''), '''
                                            || l_tab_rec.tablespace_name  || ''')',
                                      'part_maint.split_partition_short( ''' || table_name || ''', ''' || previous_part || ''', '
                                            || '''PART_' || TO_CHAR(l_target_max_part,'YYYYMMDD') || ''',to_date('''
                                            ||TO_CHAR(new_part_end_period,'yyyymmddhh24') || ''',''yyyymmddhh24''), '''
                                            || previous_tablespace_name || ''','''
                                            || l_tab_rec.tablespace_name ||''')'
                                     )
                                INTO l_local_sql
                        FROM (
                            SELECT pt.table_name table_name,
                                    previous_part.partition_name previous_part,
                                    previous_part.TABLESPACE_NAME previous_tablespace_name,
                                    '  PART_' || TO_CHAR(l_target_max_part,'yyyymmdd') new_part,
                                    Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,1) new_part_end_period
                            FROM PART_TABLES pt,
                                user_tab_partitions part_not_exist,
                                user_tab_partitions previous_part
                            WHERE pt.table_name = l_tab_rec.table_name
                            AND pt.table_name=part_not_exist.table_name(+)
                            AND pt.table_name=previous_part.table_name(+)
                            AND l_target_max_part=TO_DATE(Partition_Date_Str(part_not_exist.partition_name(+)),'yyyymmddhh24mi')
                            AND l_target_max_part<TO_DATE(Partition_Date_Str(previous_part.partition_name(+)),'yyyymmddhh24mi')
                            AND part_not_exist.partition_name IS NULL
                            ORDER BY previous_part.partition_name ASC)
                        WHERE ROWNUM=1;
                       
                        EXECUTE IMMEDIATE('begin ' || l_local_sql || '; end;');

                        DBMS_OUTPUT.PUT_LINE(' Added partition P' || TO_CHAR(l_target_max_part,'yyyymmddhh24')
                                     || ', to table '|| l_tab_rec.table_name);


                        -- if the added partition is bellow the retention period, we need to pin it.
                        IF (l_target_max_part <= l_retention_min_part)
                        THEN
                            DBMS_OUTPUT.PUT_LINE('min partition is outside of retention period:'||l_current_min_part);
                            INSERT INTO PART_PINNED
                            VALUES (l_tab_rec.part_table_id,l_target_max_part);
                                COMMIT;
                        END IF;

                        l_target_max_part:=Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,-1);

                    EXCEPTION
                        WHEN NO_DATA_FOUND THEN
                            -- THE PARTITION ALREADY EXISTS
                            l_target_max_part:=Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,-1);
                        WHEN OTHERS THEN
                             retmsg := SQLERRM;
                             l_err_msg := 'Error Adding Partition,'||l_target_max_part||' to '||l_tab_rec.table_name;
                             DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.ADD_PARTITIONS:'||l_err_msg);
                             DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||SQLERRM);
                             RAISE;
                    END;

                   ELSIF l_tab_rec.period = 'Q' THEN
                
                    BEGIN
                        SELECT DECODE(NVL(previous_part,'not_found'),'not_found',
                                      'part_maint.alter_partition_long(''' || table_name || ''',''' || new_part || ''',' 
                                            || ' to_date('''|| TO_CHAR(new_part_end_period,'yyyymmddhh24mi') || ''',''yyyymmddhh24mi''), '''
                                            || l_tab_rec.tablespace_name  || ''')',
                                      'part_maint.split_partition_long( ''' || table_name || ''', ''' || previous_part || ''',''' || new_part || ''','
                                            || ' to_date('''||TO_CHAR(new_part_end_period,'yyyymmddhh24mi') || ''',''yyyymmddhh24mi''), '''
                                            || previous_tablespace_name || ''','''
                                            || l_tab_rec.tablespace_name ||''')'
                                     )
                                INTO l_local_sql
                        FROM (
                            SELECT pt.table_name table_name,
                                    previous_part.partition_name previous_part,
                                    previous_part.TABLESPACE_NAME previous_tablespace_name,
                                    DECODE(LENGTH(Partition_Date_Str(previous_part.partition_name)),8,
                                            '  PART_' || TO_CHAR(l_target_max_part,'yyyymmdd'),
                                            '  PART_' || TO_CHAR(l_target_max_part,'yyyymmddhh24mi')) new_part,
                                    Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,1) new_part_end_period
                            FROM PART_TABLES pt,
                                user_tab_partitions part_not_exist,
                                user_tab_partitions previous_part
                            WHERE pt.table_name = l_tab_rec.table_name
                            AND pt.table_name=part_not_exist.table_name(+)
                            AND pt.table_name=previous_part.table_name(+)
                            AND l_target_max_part=TO_DATE(Partition_Date_Str(part_not_exist.partition_name(+)),'yyyymmddhh24mi')
                            AND l_target_max_part<TO_DATE(Partition_Date_Str(previous_part.partition_name(+)),'yyyymmddhh24mi')
                            AND part_not_exist.partition_name IS NULL
                            ORDER BY previous_part.partition_name ASC)
                        WHERE ROWNUM=1;
                        
                        EXECUTE IMMEDIATE('begin ' || l_local_sql || '; end;');

                        DBMS_OUTPUT.PUT_LINE(' Added partition P' || TO_CHAR(l_target_max_part,'yyyymmddhh24')
                                     || ', to table '|| l_tab_rec.table_name);


                        -- if the added partition is bellow the retention period, we need to pin it.
                        IF (l_target_max_part <= l_retention_min_part)
                        THEN
                            DBMS_OUTPUT.PUT_LINE('min partition is outside of retention period:'||l_current_min_part);
                            INSERT INTO PART_PINNED
                            VALUES (l_tab_rec.part_table_id,l_target_max_part);
                                COMMIT;
                        END IF;

                        l_target_max_part:=Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,-1);

                    EXCEPTION
                        WHEN NO_DATA_FOUND THEN
                            -- THE PARTITION ALREADY EXISTS
                            l_target_max_part:=Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,-1);
                        WHEN partition_bounds_err THEN
                            -- MERGED
                             l_err_msg := 'Error Adding Partition,'||l_target_max_part||' to '||l_tab_rec.table_name;
                             DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.ADD_PARTITIONS:'||l_err_msg);
                             DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||SQLERRM);
                            l_target_max_part:=Adm_Date.add_periods(l_target_max_part,l_tab_rec.period,-1);
                        WHEN OTHERS THEN
                             retmsg := SQLERRM;
                             l_err_msg := 'Error Adding Partition,'||l_target_max_part||' to '||l_tab_rec.table_name;
                             DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.ADD_PARTITIONS:'||l_err_msg);
                             DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||SQLERRM);
                             RAISE;
                    END;
                                    
                   END IF;
                
               END LOOP;

           END LOOP;

       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   END add_partitions;

   -------------------------------------------------------------------
   -- delete_paritions - Deletes paritions that are no longer used.
   --------------------------------------------------------------------`
   PROCEDURE delete_partitions(l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                            l_sub_type IN VARCHAR2 DEFAULT NULL,l_tab_type IN VARCHAR2 DEFAULT NULL,
                            l_filter_name IN VARCHAR2 DEFAULT NULL)
   IS

      CURSOR cur_table_period (l_sdate_in DATE, l_edate_in DATE,
                               l_type_in VARCHAR2, l_sub_type_in VARCHAR2 DEFAULT NULL,
                               l_tab_type_in VARCHAR2 DEFAULT NULL, l_filter_name_in VARCHAR2 DEFAULT NULL)  IS
         SELECT v.period
         FROM PART_MAINT_PART_SUMMARY_VW v
         WHERE UPPER(v.TYPE) LIKE NVL(UPPER(l_type_in),'%')
         AND UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name_in),'%');
                                         
      CURSOR cur_table_details(l_sdate_in DATE, l_edate_in DATE,
                               l_type_in VARCHAR2, l_sub_type_in VARCHAR2 DEFAULT NULL,
                               l_tab_type_in VARCHAR2 DEFAULT NULL, l_filter_name_in VARCHAR2 DEFAULT NULL)  IS
         SELECT v.table_name, utp.partition_name
         FROM PART_MAINT_PART_SUMMARY_VW v, user_tab_partitions utp
         WHERE utp.table_name = v.table_name
         AND UPPER(v.TYPE) LIKE NVL(UPPER(l_type_in),'%')
         AND UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name_in),'%')
         AND utp.partition_name >= Part_Maint.partition_name(l_sdate_in)
         AND utp.partition_name <= Part_Maint.partition_name(l_edate_in)
         MINUS
         (SELECT v.table_name, utp.partition_name
         FROM PART_MAINT_PART_SUMMARY_VW v, user_tab_partitions utp, PART_PINNED pp
         WHERE utp.table_name = v.table_name
         AND pp.part_table_id = v.part_table_id
         AND UPPER(v.TYPE)=UPPER(l_type_in)
         AND UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name_in),'%')
         AND utp.partition_name >= Part_Maint.partition_name(l_sdate_in)
         AND utp.partition_name <= Part_Maint.partition_name(l_edate_in)
         AND pp.part_eff_date = Part_Maint.partition_date(utp.partition_name));

      CURSOR cur_table_qdetails(l_sdate_in DATE, l_edate_in DATE,
                               l_type_in VARCHAR2, l_sub_type_in VARCHAR2 DEFAULT NULL,
                               l_tab_type_in VARCHAR2 DEFAULT NULL, l_filter_name_in VARCHAR2 DEFAULT NULL)  IS
         SELECT v.table_name, utp.partition_name
         FROM PART_MAINT_PART_SUMMARY_VW v, user_tab_partitions utp
         WHERE utp.table_name = v.table_name
         AND UPPER(v.TYPE) LIKE NVL(UPPER(l_type_in),'%')
         AND UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name_in),'%')
         AND utp.partition_name >= Part_Maint.partition_name(l_sdate_in, 'L')
         AND utp.partition_name <= Part_Maint.partition_name(l_edate_in, 'L')
         MINUS
         (SELECT v.table_name, utp.partition_name
         FROM PART_MAINT_PART_SUMMARY_VW v, user_tab_partitions utp, PART_PINNED pp
         WHERE utp.table_name = v.table_name
         AND pp.part_table_id = v.part_table_id
         AND UPPER(v.TYPE)=UPPER(l_type_in)
         AND UPPER(v.sub_type) LIKE NVL(UPPER(l_sub_type_in),'%')
         AND UPPER(v.table_type) LIKE NVL(UPPER(l_tab_type_in),'%')
         AND UPPER(v.table_name) LIKE NVL(UPPER(l_filter_name_in),'%')
         AND utp.partition_name >= Part_Maint.partition_name(l_sdate_in, 'L')
         AND utp.partition_name <= Part_Maint.partition_name(l_edate_in, 'L')
         AND pp.part_eff_date = Part_Maint.partition_date(utp.partition_name));
         
      invalid_date_err EXCEPTION;

      resource_busy EXCEPTION;
      PRAGMA EXCEPTION_INIT( resource_busy, -00054 );

      l_sdate          DATE;
      l_edate          DATE;
      retval           BOOLEAN        := TRUE;
      needs_attempt    BOOLEAN        := TRUE;
      attempt_count    INTEGER        := 0;
      l_part_table_id  INTEGER        := 0;
      retmsg           VARCHAR2(9000) := NULL;
      vprc             VARCHAR2(30)   := 'DeletePartition';
      l_str            VARCHAR2 (1000);
      l_err_msg        VARCHAR2(200)  := NULL;


   BEGIN

       --Check for the validity of the dates
       BEGIN
           SELECT TO_DATE(l_sdate_in,'yyyymmddHH24'),TO_DATE(l_edate_in,'yyyymmddHH24')
           INTO l_sdate,l_edate
           FROM dual;
       EXCEPTION WHEN OTHERS THEN
           RAISE invalid_date_err;
       END;

       IF ( tabletypeexists( l_type, l_sub_type))
       THEN
         
        FOR l_period_rec IN cur_table_period(l_sdate, l_edate, l_type, l_sub_type, l_tab_type, l_filter_name) LOOP
         IF NOT l_period_rec.period = 'Q' THEN
         
           FOR l_tab_rec IN cur_table_details(l_sdate, l_edate, l_type, l_sub_type, l_tab_type, l_filter_name) LOOP

               BEGIN

                   IF (partition_exists( l_tab_rec.table_name, l_tab_rec.partition_name))
                   THEN
                       l_str := 'ALTER TABLE '
                                || l_tab_rec.table_name
                                || ' DROP PARTITION ' || l_tab_rec.partition_name;

                       needs_attempt := TRUE;

                       WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) )
                       LOOP
                           BEGIN
                            -- Lock the partition
                            lock_table(l_tab_rec.table_name, l_tab_rec.partition_name);
                            EXECUTE IMMEDIATE l_str;
                            needs_attempt := FALSE;

                           EXCEPTION
                              WHEN resource_busy THEN
                                 attempt_count := attempt_count + 1;
                                 dbms_application_info.set_client_info( l_tab_rec.table_name ||
                                    ': retry ' || TO_CHAR(attempt_count) );
                                 dbms_lock.sleep(g_resource_interval);
                           END;
                       END LOOP;

                   END IF;

                   IF (retval) THEN
                          DBMS_OUTPUT.PUT_LINE('PART_MAINT: Deletion of partition '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name||' done');
                   ELSE
                       DBMS_OUTPUT.PUT_LINE('PART_MAINT: Deletion of partition '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name||' failed. '||SUBSTR(retmsg,1,200));
                   END IF;

               EXCEPTION WHEN OTHERS THEN
                      IF retval THEN
                        retmsg := SQLERRM;
                      END IF;
                      retval := FALSE;

                      l_err_msg := 'Error Deleting Partitions, for '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name;
                      DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.DROP_PARTITIONS:'||l_err_msg);
                      DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||retmsg);
                      retval := FALSE;
                      RAISE;
               END;

           END LOOP; -- End of FOR LOOP
           
         ELSIF l_period_rec.period = 'Q' THEN
         
           FOR l_tab_rec IN cur_table_qdetails(l_sdate, l_edate, l_type, l_sub_type, l_tab_type, l_filter_name) LOOP

               BEGIN

                   IF (partition_exists( l_tab_rec.table_name, l_tab_rec.partition_name))
                   THEN
                       l_str := 'ALTER TABLE '
                                || l_tab_rec.table_name
                                || ' DROP PARTITION ' || l_tab_rec.partition_name;

                       needs_attempt := TRUE;

                       WHILE ( needs_attempt AND ( attempt_count < g_resource_attempts ) )
                       LOOP
                           BEGIN
                            -- Lock the partition
                            lock_table(l_tab_rec.table_name, l_tab_rec.partition_name);
                            EXECUTE IMMEDIATE l_str;
                            needs_attempt := FALSE;

                           EXCEPTION
                              WHEN resource_busy THEN
                                 attempt_count := attempt_count + 1;
                                 dbms_application_info.set_client_info( l_tab_rec.table_name ||
                                    ': retry ' || TO_CHAR(attempt_count) );
                                 dbms_lock.sleep(g_resource_interval);
                           END;
                       END LOOP;

                   END IF;

                   IF (retval) THEN
                          DBMS_OUTPUT.PUT_LINE('PART_MAINT: Deletion of partition '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name||' done');
                   ELSE
                       DBMS_OUTPUT.PUT_LINE('PART_MAINT: Deletion of partition '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name||' failed. '||SUBSTR(retmsg,1,200));
                   END IF;

               EXCEPTION WHEN OTHERS THEN
                      IF retval THEN
                        retmsg := SQLERRM;
                      END IF;
                      retval := FALSE;

                      l_err_msg := 'Error Deleting Partitions, for '||l_tab_rec.table_name||'.'||l_tab_rec.partition_name;
                      DBMS_OUTPUT.PUT_LINE('ERROR: PART_MAINT.DROP_PARTITIONS:'||l_err_msg);
                      DBMS_OUTPUT.PUT_LINE(SQLCODE||', '||retmsg);
                      retval := FALSE;
                      RAISE;
               END;


           END LOOP; -- End of WHILE LOOP

         END IF;
        END LOOP;
       ELSE
           DBMS_OUTPUT.PUT_LINE(' Table Type '|| l_type ||', '||l_sub_type||' is not a valid Type');
           DBMS_OUTPUT.PUT_LINE(' Check for valid tabletypes with the part_admin -listtypes option');
       END IF;

   END delete_partitions;

END Part_Maint;
/
SHOW ERROR PACKAGE BODY Part_Maint

