/*********************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 * 
 * --SA4.1_PID--
 * Copyright IBM Corp. 2005, 2006, 2007, 2008
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what
 * been deposited with the U.S. Copyright Office. 
 *-----------------------------------------------                
 * Author: Thierry Supplisson
 * Created: 06/01/2005
 * Description:
 *   Script to create the admin packages
 *-----------------------------------------------                
 * MODIFICATION HISTORY
 * DATE        MODIFIED BY    MODIFICATION
 ********************************************************************/

CREATE OR REPLACE PACKAGE
-- -----------------------------------------------------------------------
--
-- Copyright IBM Corp. 2007
--
-- @(#) part_maintPack.sql 1.10.0 07/03/23 11:02:31
-- @(#)
--
-- -----------------------------------------------------------------------
  part_maint IS

     TYPE taskRec IS RECORD (
        JOB_ID           INTEGER,
        TASK_ID          INTEGER,
        STATUS_CODE      INTEGER,
        OPERATION        INTEGER,
        TASK_DATE        DATE,
        PART_TABLE_ID    INTEGER,
        TABLE_NAME       VARCHAR2(30),
        TABLESPACE_NAME  VARCHAR2(30),
        PERIOD           CHAR(1),
        DATA_RETENTION   INTEGER,
        PAST_RETENTION   INTEGER,
        FUTURE_RETENTION INTEGER,
        NEEDS_MOVE       CHAR(1),
        NEEDS_EXPORT     CHAR(1),
        NEEDS_PURGE      CHAR(1),
        NEEDS_ANALYZE    CHAR(1),
        ANALYZE_MASK     CHAR(7),
        PART_NAME        VARCHAR2(30)
     );

   TYPE TbsSpace is table of number index by VARCHAR2(30);

   TYPE t_ref_cursor IS REF CURSOR;

   g_TbsSpace TbsSpace;

   PROCEDURE maintainminly;

   PROCEDURE maintainhourly;

   PROCEDURE maintaindaily;

   PROCEDURE maintainweekly;

   PROCEDURE maintainmonthly;

   FUNCTION partition_date(partition_name VARCHAR2)
			   RETURN DATE PARALLEL_ENABLE;
   PRAGMA RESTRICT_REFERENCES (partition_date, WNDS, WNPS, RNDS, RNPS);

   FUNCTION partition_date(partition_name VARCHAR2, p_mode CHAR)
			   RETURN DATE PARALLEL_ENABLE;
   PRAGMA RESTRICT_REFERENCES (partition_date, WNDS, WNPS, RNDS, RNPS);

   FUNCTION partition_name(partition_date DATE)
			   RETURN VARCHAR2 PARALLEL_ENABLE;
   PRAGMA RESTRICT_REFERENCES (partition_name, WNDS, WNPS, RNDS, RNPS);

   FUNCTION partition_name(partition_date DATE, p_mode CHAR)
			   RETURN VARCHAR2 PARALLEL_ENABLE;
   PRAGMA RESTRICT_REFERENCES (partition_name, WNDS, WNPS, RNDS, RNPS);

   PROCEDURE lock_table(table_name VARCHAR2, partition_name VARCHAR2 DEFAULT NULL);

   FUNCTION partition_exists(table_name_in VARCHAR2,
			   partition_name_in VARCHAR2) RETURN BOOLEAN;

   PROCEDURE part_session( p_orajobid IN INTEGER, p_jobid IN INTEGER, p_taskid IN INTEGER);

   PROCEDURE get_pm_sessions( results OUT t_ref_cursor );

   PROCEDURE update_pm_sessions( l_value IN VARCHAR2);
   
   PROCEDURE get_valid_types( l_type IN VARCHAR2, l_subtype IN VARCHAR2, results OUT t_ref_cursor );

   PROCEDURE get_main_types( l_type IN VARCHAR2, results OUT t_ref_cursor );

   PROCEDURE get_space_usage( results OUT t_ref_cursor );
   
   FUNCTION tabletypeexists(type_name_in VARCHAR2,
			   subtype_name_in VARCHAR2 DEFAULT NULL) RETURN BOOLEAN;
               
   FUNCTION CheckSpace(table_name_in VARCHAR2, tablespace_name_in VARCHAR2, data_retention_in INTEGER,
			   future_retention_in INTEGER, checkparam VARCHAR2) RETURN INTEGER;

   PROCEDURE get_list_partitions(l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                             l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                             l_sdate_in IN VARCHAR2 DEFAULT NULL, l_edate_in IN VARCHAR2 DEFAULT NULL,
                             results OUT t_ref_cursor);

   PROCEDURE get_summary_partitions( l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                                  l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                                  results OUT t_ref_cursor );

   PROCEDURE get_pinned_partitions( l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                              l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                              l_sdate_in IN VARCHAR2 DEFAULT NULL, l_edate_in IN VARCHAR2 DEFAULT NULL,
                              results OUT t_ref_cursor );

   PROCEDURE get_parameters (l_type IN VARCHAR2, l_sub_type IN VARCHAR2 DEFAULT NULL,
                                  l_tab_type IN VARCHAR2 DEFAULT NULL, l_filter_name IN VARCHAR2 DEFAULT NULL,
                                  results OUT t_ref_cursor );

   PROCEDURE get_default_parameters ( results OUT t_ref_cursor );

   PROCEDURE get_analyze_mask_details ( results OUT t_ref_cursor );

   PROCEDURE update_parameters (l_param IN VARCHAR2, l_value IN VARCHAR2, l_type IN VARCHAR2,
                                l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                                l_filter_name IN VARCHAR2 DEFAULT NULL);

   PROCEDURE get_pm_status(results OUT t_ref_cursor);

   PROCEDURE get_pm_logs(results OUT t_ref_cursor);

   PROCEDURE get_pm_latest_logs(results OUT t_ref_cursor);

   PROCEDURE get_pm_detailed_logs(l_log_id IN VARCHAR2 DEFAULT NULL, results OUT t_ref_cursor);

   PROCEDURE get_pm_error_logs(results OUT t_ref_cursor);

   PROCEDURE add_partitions (l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                                l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                                l_filter_name IN VARCHAR2 DEFAULT NULL);

   PROCEDURE delete_partitions (l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                               l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                               l_filter_name IN VARCHAR2 DEFAULT NULL);

   PROCEDURE pin_partitions (l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                                l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                                l_filter_name IN VARCHAR2 DEFAULT NULL);

   PROCEDURE unpin_partitions (l_sdate_in IN VARCHAR2, l_edate_in IN VARCHAR2, l_type IN VARCHAR2,
                                l_sub_type IN VARCHAR2 DEFAULT NULL, l_tab_type IN VARCHAR2 DEFAULT NULL,
                                l_filter_name IN VARCHAR2 DEFAULT NULL);

   procedure alter_partition_short(p_table_name varchar2, p_new_part varchar2, p_new_part_end_period date,
                            p_tablespace_name varchar2);

   procedure alter_partition_long(p_table_name varchar2, p_new_part varchar2,  p_new_part_end_period date,
                            p_tablespace_name varchar2);
                            
   procedure split_partition_short(p_table_name varchar2, p_previous_part varchar2,
                            p_new_part varchar2, p_new_part_end_period date,
                            p_previous_tablespace_name varchar2, p_tablespace_name varchar2);

   procedure split_partition_long(p_table_name varchar2, p_previous_part varchar2,
                            p_new_part varchar2, p_new_part_end_period date,
                            p_previous_tablespace_name varchar2, p_tablespace_name varchar2);
END part_maint;
/
