#!/bin/ksh

WAR_FILE=customer_care.war
CMD=$(basename $0)
INSTALL_LOC=$(dirname $0)/../
LIBDIR=${INSTALL_LOC}/lib
LIBFILE=${LIBDIR}/flcc_functions.kshlib
FILE_LOCATION=$(dirname $0)/../
FLCC_LOGS=$TIP_HOME/logs/customer_care
LOGFILE=$TIP_HOME/logs/customer_care/configure_flcc.log
INSTALL=NO
REMOVE=NO

. ${LIBFILE}
if [ $? -ne 0 ]
then
  echo "ERROR - unable to source ${LIBFILE}."
  exit 1
fi

while [ $# != 0 ]
do
  case "$1" in

        -profile)
               shift
               if [ $# = 0 ]
               then
                 echo "${CMD} [${PID}]: ERROR: TIP Profile expected"
                 usage
                 exit 1
               fi
               export TIP_PROFILE=$1
               ;;
        -user)
               shift
               if [ $# = 0 ]
               then
                 echo "${CMD} [${PID}]: ERROR: TIP User expected"
                 usage
                 exit 1
               fi
               export TIP_USER=$1
               ;;
        -passwd)
               shift
               if [ $# = 0 ]
               then
                 echo "${CMD} [${PID}]: ERROR: TIP Password expected"
                 usage
                 exit 1
               fi
               export TIP_PASSWD=$1
               ;;
               
        -mode)
               shift
               case $1 in
                 install)
                          INSTALL=YES
                          ;;
                  remove)
                          REMOVE=YES
                          ;;
                       *)
                          echo "${CMD} [${PID}]: ERROR: install or remove directive expected"
                          usage
                          ;;
               esac
               ;;

             *)
               usage
               ;;
  esac
  shift
done


usage()
{
# echo usage statement here
  echo "${CMD}: [${PID}]: Usage: ${CMD} -profile <TIP Profile> -user <TIP Admin User> -passwd <TIP Admin Password> -mode <install|remove>"
  return 0

}

create_logging()
{
  mkdir -p $FLCC_LOGS 
}

deploy_components()
{

  echo "Deploying Components... \c"
  # copy start script to $TIP_HOME/bin
  cp $FILE_LOCATION/bin/startflcc.sh $TIP_HOME/bin
  chmod 755 $TIP_HOME/bin/startflcc.sh
  
  create_logging
  
  # extract war file to $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear
  mkdir -p $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war
  cp $FILE_LOCATION/${WAR_FILE} $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war
  cd $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war
  jar -xf $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war/customer_care.war
  rm $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war/customer_care.war
  echo "done."
  
  FIRST_LINE_ROLE="firstline"  
  SECOND_LINE_ROLE="secondline"  
  VIP_ROLE="vip"  
  
       	echo "Enter First Line Role Name (${FIRST_LINE_ROLE}): "
     	read SOMEVAR
     	[ -z "${SOMEVAR}" ] || FIRST_LINE_ROLE=${SOMEVAR}  
     	
     	echo "Enter Second Line Role Name (${SECOND_LINE_ROLE}): "
     	read SOMEVAR
     	[ -z "${SOMEVAR}" ] || SECOND_LINE_ROLE=${SOMEVAR}  
     	
     	echo "Enter VIP Role Name (${VIP_ROLE}): "
     	read SOMEVAR
     	[ -z "${SOMEVAR}" ] || VIP_ROLE=${SOMEVAR}  
     	
       	cat >$TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war/WEB-INF/web.xml <<EOF
<?xml version="1.0" encoding="UTF-8"?>

<!DOCTYPE web-app PUBLIC "-//Sun Microsystems, Inc.//DTD Web Application 2.3//EN" "http://java.sun.com/dtd/web-app_2_3.dtd">

<web-app>
    <display-name>Customer Care Component</display-name>
    
    <description>Customer Care Component</description>
    
    <servlet>
        <servlet-name>CustomerSearch</servlet-name>
        <servlet-class>com.ibm.sa.customercare.servlet.CustomerSearchServlet</servlet-class>
      <init-param>
            <param-name>vip.role.name</param-name>
            <param-value>${VIP_ROLE}</param-value>
      </init-param>
            <init-param>
            <param-name>customer.type.attr</param-name>
            <param-value>customer_type</param-value>
      </init-param>
     <init-param>
            <param-name>customer.type.vip</param-name>
            <param-value>VVIP</param-value>
      </init-param>
               
    </servlet>
    
    <servlet>
        <servlet-name>ReportLaunch</servlet-name>
        <servlet-class>com.ibm.sa.customercare.servlet.ReportLaunchServlet</servlet-class>
         
      <init-param>
            <param-name>firstline.role.name</param-name>
            <param-value>${FIRST_LINE_ROLE}</param-value>
      </init-param>
      
      <init-param>
            <param-name>secondline.role.name</param-name>
            <param-value>${SECOND_LINE_ROLE}</param-value>
      </init-param>
      <init-param>
            <param-name>vip.role.name</param-name>
            <param-value>${VIP_ROLE}</param-value>
      </init-param>
    </servlet>
    
    <servlet-mapping>
        <servlet-name>CustomerSearch</servlet-name>
        <url-pattern>/customerSearch</url-pattern>
    </servlet-mapping>
    
    <servlet-mapping>
        <servlet-name>ReportLaunch</servlet-name>
        <url-pattern>/reportLaunch</url-pattern>
    </servlet-mapping>
    
    <session-config>
        <session-timeout>1000</session-timeout>
    </session-config>
    
    <security-constraint>
        <web-resource-collection>
            <web-resource-name>FLCC Protected Area</web-resource-name>
            <url-pattern>/*</url-pattern>
            <http-method>GET</http-method>
            <http-method>POST</http-method>
        </web-resource-collection>
        <auth-constraint>
            <role-name>iscadmins</role-name>
            <role-name>tcrPortalOperator</role-name> 
        </auth-constraint>
    </security-constraint>
    <login-config>
        <auth-method>BASIC</auth-method>
    </login-config>
    <security-role>
            <role-name>iscadmins</role-name>
    </security-role>
</web-app>

EOF

}

remove_components()
{

  echo "Removing Components... \c"
  # remove start script
  rm $TIP_HOME/bin/startflcc.sh
  
  # Remove logging area
  rm -rf $FLCC_LOGS 
  
  # Remove web-app
  rm -rf $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/customer_care.war
  echo "done."
}

was_install_config()
{
	#Configure web-app
	cd -
	echo "Updating ${TIP_PROFILE} configuration... \c"
    ${TIP_HOME}/bin/wsadmin.sh -profileName ${TIP_PROFILE} \
          -username ${TIP_USER} -password ${TIP_PASSWD} <<EOF >>$LOGFILE
       \$AdminApp update isc modulefile {-operation add -contents $TIP_HOME/profiles/TIPProfile/installedApps/TIPCell/isc.ear/${WAR_FILE} -contenturi ${WAR_FILE} -contextroot /ibm/customer_care -usedefaultbindings -custom paavalidation=false -MapWebModToVH {{.* .* default_host}}}
       \$AdminConfig save
EOF
     [ $? -ne 0 ] && echo "ERROR: Update of ${TIP_PROFILE} configuration failed." && exit 1
     echo "done."
     echo "Updating security configuration...\c"
     echo  "\nCTAUTHENTICATION {\
com.comnitel.jaas.ldap.LdapLoginModule optional;\
};" >> ${TIP_HOME}/profiles/TIPProfile/properties/wsjaas.conf
     echo "done."    
  return 0
	
}

was_uninstall_config()
{		
	#Deconfigure web-app
     echo "Updating ${TIP_PROFILE} configuration... \c"
     ${TIP_HOME}/bin/wsadmin.sh -profileName ${TIP_PROFILE} \
          -username ${TIP_USER} -password ${TIP_PASSWD} <<EOF >>$LOGFILE
       \$AdminApp update isc modulefile {-operation delete -contenturi customer_care.war}
       \$AdminConfig save
EOF
    [ $? -ne 0 ] && echo "ERROR: Update of ${TIP_PROFILE} configuration failed." && exit 1
     echo "done."
}

check_env()
{
    # Check TIP_HOME is set
    if [ "X$TIP_HOME" = "X" ]; then
        echo "ERROR: TIP_HOME variable is mandatory"
        usage
        exit 1
    fi

    # Check the flcc war file is present
    if [ ! -f $FILE_LOCATION/${WAR_FILE} ]; then
       echo "ERROR: Web archive does not exist : $FILE_LOCATION/${WAR_FILE}"
       exit 1
    fi
    
    # Check TIP_HOME set correctly
    if [ ! -f $TIP_HOME/bin/startServer.sh ]; then
       echo "ERROR: TIP_HOME variable incorrectly set : $TIP_HOME"
       exit 1
    fi
}

restart_tip() {

  echo "Restarting the TIP Server..."

  $TIP_HOME/bin/stopServer.sh server1 -username $TIP_USER -password $TIP_PASSWD -profileName $TIP_PROFILE
  $TIP_HOME/bin/startServer.sh server1 -username $TIP_USER -password $TIP_PASSWD -profileName $TIP_PROFILE

  echo "OK."

  return 0

} # restart_tip


# FLCC Install procedure

if [ "${INSTALL}" = "NO"  -a "${REMOVE}" = "NO" ]
then
  echo "${CMD} [${PID}]: ERROR: Must specify a mode (install or remove)"
  usage
  exit 1
fi

if [ "${INSTALL}" = "YES" ]
then
  check_env
  deploy_components
  was_install_config
  restart_tip
fi

if [ "${REMOVE}" = "YES" ]
then
  check_env
  create_logging
  was_uninstall_config
  remove_components
  restart_tip
fi


